import os
from collections import OrderedDict

from astropy.wcs.utils import wcs_to_celestial_frame
from astropy.coordinates import BaseRADecFrame

from qtpy import QtWidgets, QtCore, QtGui
from qtpy.QtWidgets import QMenu, QAction, QInputDialog, QActionGroup

from glue.utils.qt import load_ui
from glue.utils.qt import get_qapp
from glue.config import qt_fixed_layout_tab
from glue.external.echo import keep_in_sync, SelectionCallbackProperty
from glue.external.echo.qt.connect import connect_combo_selection, UserDataWrapper, _find_combo_data, _find_combo_text
from glue.core.data_combo_helper import ComponentIDComboHelper
from glue.core.message import (SettingsChangeMessage, SubsetUpdateMessage,
                               SubsetDeleteMessage, EditSubsetMessage)
from glue.utils.matplotlib import freeze_margins
from glue.dialogs.component_arithmetic.qt import ArithmeticEditorWidget

from specviz.third_party.glue.data_viewer import SpecVizViewer
from specviz.core.events import dispatch

from .toolbar import CubevizToolbar
from .image_viewer import CubevizImageViewer

from .controls.slice import SliceController
from .controls.overlay import OverlayController

from .controls.flux_units import FluxUnitController
from .controls.wavelengths import WavelengthController
from .tools import collapse_cube
from .tools import moment_maps, smoothing
from .tools.wavelengths_ui import WavelengthUI
from .tools.spectral_operations import SpectralOperationHandler


DEFAULT_NUM_SPLIT_VIEWERS = 3
DEFAULT_TOOLBAR_ICON_SIZE = 18


class WidgetWrapper(QtWidgets.QFrame):

    def __init__(self, widget=None, tab_widget=None, toolbar=False, parent=None):
        super(WidgetWrapper, self).__init__(parent=parent)
        self.tab_widget = tab_widget
        self._widget = widget

        self.layout = QtWidgets.QVBoxLayout()
        self.layout.setSpacing(0)
        self.layout.setContentsMargins(0, 0, 0, 0)

        if toolbar:
            self.setObjectName('widgetWrapper')
            self.setStyleSheet(
                '#widgetWrapper { border: 2px solid #aaa; border-radius: 2px }')
            self.create_toolbar()

        # Add the image viewer itself to the layout
        self.layout.addWidget(widget)
        self.setLayout(self.layout)

        if toolbar:
            self.create_stats()

    def create_toolbar(self):
        self.tb = QtWidgets.QToolBar()

        self.combo = QtWidgets.QComboBox(enabled=False)
        self.tb.addWidget(self.combo)

        self.checkbox = QtWidgets.QCheckBox('Synced', enabled=False, checked=True)
        self.tb.addWidget(self.checkbox)

        # Add a spacer so that the slice text is right-aligned
        self.spacer = QtWidgets.QWidget()
        self.spacer.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self.tb.addWidget(self.spacer)

        self.slice_text = QtWidgets.QLabel('')
        self.tb.addWidget(self.slice_text)

        self.layout.addWidget(self.tb)

    def create_stats(self):

        self.stats_widget = QtWidgets.QWidget()
        self.stats_layout = QtWidgets.QVBoxLayout()
        self.stats_layout.setSpacing(0)
        self.stats_layout.setContentsMargins(0, 0, 0, 0)
        self.stats_widget.setLayout(self.stats_layout)
        self.layout.addWidget(self.stats_widget)

        bold_font = QtGui.QFont()
        bold_font.setBold(True)

        self.stats_label = QtWidgets.QLabel('', font=bold_font)
        self.stats_layout.addWidget(self.stats_label)

        self.stats_text = QtWidgets.QLabel()
        self.stats_layout.addWidget(self.stats_text)

    def set_stats_visible(self, visible):
        self.stats_widget.setVisible(visible)

    def set_stats_text(self, label, text):
        self.stats_label.setText(label)
        self.stats_text.setText(text)

    def widget(self):
        return self._widget


@qt_fixed_layout_tab
class CubeVizLayout(QtWidgets.QWidget):
    """
    The 'CubeViz' layout, with three image viewers and one spectrum viewer.
    """

    LABEL = "CubeViz"
    subWindowActivated = QtCore.Signal(object)

    single_viewer_attribute = SelectionCallbackProperty(default_index=0)
    viewer1_attribute = SelectionCallbackProperty(default_index=0)
    viewer2_attribute = SelectionCallbackProperty(default_index=1)
    viewer3_attribute = SelectionCallbackProperty(default_index=2)

    def __init__(self, session=None, parent=None):
        super(CubeVizLayout, self).__init__(parent=parent)

        self._cubeviz_toolbar = None

        if not getattr(session.application, '_has_cubeviz_toolbar', False):
            self._cubeviz_toolbar = CubevizToolbar(application=session.application)
            session.application.insertToolBar(session.application._data_toolbar,
                                              self._cubeviz_toolbar)
            session.application._has_cubeviz_toolbar = True

        self.session = session
        self._has_data = False
        self._option_buttons = []

        self._data = None

        self.ui = load_ui('layout.ui', self,
                          directory=os.path.dirname(__file__))

        self.cube_views = []

        # Create the cube viewers and register to the hub.
        for _ in range(DEFAULT_NUM_SPLIT_VIEWERS + 1):
            ww = WidgetWrapper(CubevizImageViewer(
                    self.session, cubeviz_layout=self), tab_widget=self,
                    toolbar=True)
            self.cube_views.append(ww)
            ww._widget.register_to_hub(self.session.hub)

        self.set_toolbar_icon_size(DEFAULT_TOOLBAR_ICON_SIZE)

        # Create specviz viewer and register to the hub.
        self.specviz = WidgetWrapper(
            SpecVizViewer(self.session, layout=self), tab_widget=self, toolbar=False)
        self.specviz._widget.register_to_hub(self.session.hub)

        self.single_view = self.cube_views[0]
        self.split_views = self.cube_views[1:]

        self._synced_checkboxes = [view.checkbox for view in self.cube_views]

        for view, checkbox in zip(self.cube_views, self._synced_checkboxes):
            view._widget.assign_synced_checkbox(checkbox)

        # Add the views to the layouts.
        self.ui.single_image_layout.addWidget(self.single_view)
        for viewer in self.split_views:
            self.ui.image_row_layout.addWidget(viewer)

        self.ui.specviz_layout.addWidget(self.specviz)

        self.subWindowActivated.connect(self._update_active_view)

        self.ui.sync_button.clicked.connect(self._on_sync_click)
        self.ui.button_toggle_image_mode.clicked.connect(
            self._toggle_image_mode)

        # This is a list of helpers for the viewer combo boxes. New data
        # collections should be added to each helper in this list using the
        # ``append_data`` method to ensure that the new data components are
        # populated into the combo boxes.
        self._viewer_combo_helpers = []

        # This tracks the current positions of cube viewer axes when they are hidden
        self._viewer_axes_positions = []

        # Indicates whether cube viewer toolbars are currently visible or not
        self._toolbars_visible = True

        # Indicates whether subset stats should be displayed or not
        self._stats_visible = True

        self._slice_controller = SliceController(self)
        self._overlay_controller = OverlayController(self)

        self._wavelength_controller = WavelengthController(self)
        self._flux_unit_controller = FluxUnitController(self)

        # Add menu buttons to the cubeviz toolbar.
        self.ra_dec_format_menu = None
        self._init_menu_buttons()

        self.sync = {}

        app = get_qapp()
        app.installEventFilter(self)
        self._last_click = None
        self._active_view = None
        self._active_cube = None
        self._last_active_view = None
        self._active_split_cube = None

        # Set the default to parallel image viewer
        self._single_viewer_mode = False
        self.ui.button_toggle_image_mode.setText('Single Image Viewer')

        # Add this class to the specviz dispatcher watcher
        dispatch.setup(self)

    def _init_menu_buttons(self):
        """
        Add the two menu buttons to the tool bar. Currently two are defined:
            View - for changing the view of the active window
            Data Processing - for applying a data processing step to the data.

        :return:
        """
        self._option_buttons = [
            self.ui.view_option_button,
            self.ui.cube_option_button
        ]

        # Create the View Menu
        view_menu = self._dict_to_menu(OrderedDict([
            ('Hide Axes', ['checkable', self._toggle_viewer_axes]),
            ('Hide Toolbars', ['checkable', self._toggle_toolbars]),
            ('Hide Spaxel Value Tooltip', ['checkable', self._toggle_hover_value]),
            ('Hide Stats', ['checkable', self._toggle_stats_display]),
            ('Convert Flux Units', lambda: self._open_dialog('Convert Flux Units', None)),
            ('Wavelength Units/Redshift', lambda: self._open_dialog('Wavelength Units/Redshift', None))
        ]))

        # Add toggle RA-DEC format:
        format_menu = view_menu.addMenu("RA-DEC Format")
        format_action_group = QActionGroup(format_menu)
        self.ra_dec_format_menu = format_menu

        # Make sure to change all instances of the the names
        # of the formats if modifications are made to them.
        for format_name in ["Sexagesimal", "Decimal Degrees"]:
            act = QAction(format_name, format_menu)
            act.triggered.connect(self._toggle_all_coords_in_degrees)
            act.setActionGroup(format_action_group)
            act.setCheckable(True)
            act.setChecked(True) if format == "Sexagesimal" else act.setChecked(False)
            format_menu.addAction(act)

        self.ui.view_option_button.setMenu(view_menu)

        # Create the Data Processing Menu
        cube_menu = self._dict_to_menu(OrderedDict([
            ('Collapse Cube', lambda: self._open_dialog('Collapse Cube', None)),
            ('Spatial Smoothing', lambda: self._open_dialog('Spatial Smoothing', None)),
            ('Moment Maps', lambda: self._open_dialog('Moment Maps', None)),
            ('Arithmetic Operations', lambda: self._open_dialog('Arithmetic Operations', None))
        ]))
        self.ui.cube_option_button.setMenu(cube_menu)

    def _dict_to_menu(self, menu_dict):
        '''Stolen shamelessly from specviz. Thanks!'''
        menu_widget = QMenu()
        for k, v in menu_dict.items():
            if isinstance(v, dict):
                new_menu = menu_widget.addMenu(k)
                self._dict_to_menu(v, menu_widget=new_menu)
            else:
                act = QAction(k, menu_widget)

                if isinstance(v, list):
                    if v[0] == 'checkable':
                        v = v[1]
                        act.setCheckable(True)
                        act.setChecked(False)

                act.triggered.connect(v)
                menu_widget.addAction(act)
        return menu_widget

    def set_toolbar_icon_size(self, size):
        for view in self.cube_views:
            view._widget.toolbar.setIconSize(QtCore.QSize(size, size))

    def handle_settings_change(self, message):
        if isinstance(message, SettingsChangeMessage):
            self._slice_controller.update_index(self.synced_index)

    @property
    def synced_index(self):
        return self._slice_controller.synced_index

    def handle_subset_action(self, message):
        self.refresh_viewer_combo_helpers()
        if isinstance(message, SubsetUpdateMessage):
            for combo, viewer in zip(self._viewer_combo_helpers, self.cube_views):
                viewer._widget.show_roi_stats(combo.selection, message.subset)
        elif isinstance(message, SubsetDeleteMessage):
            for viewer in self.cube_views:
                viewer._widget.show_slice_stats()
        elif isinstance(message, EditSubsetMessage) and not message.subset:
            for viewer in self.cube_views:
                viewer._widget.show_slice_stats()

    def _set_pos_and_margin(self, axes, pos, marg):
        axes.set_position(pos)
        freeze_margins(axes, marg)

    def _hide_viewer_axes(self):
        for viewer in self.cube_views:
            viewer._widget.toggle_hidden_axes(True)
            axes = viewer._widget.axes
            # Save current axes position and margins so they can be restored
            pos = axes.get_position(), axes.resizer.margins
            self._viewer_axes_positions.append(pos)
            self._set_pos_and_margin(axes, [0, 0, 1, 1], [0, 0, 0, 0])
            viewer._widget.figure.canvas.draw()

    def _toggle_viewer_axes(self):
        # If axes are currently hidden, restore the original positions
        if self._viewer_axes_positions:
            for viewer, pos in zip(self.cube_views, self._viewer_axes_positions):
                viewer._widget.toggle_hidden_axes(False)
                axes = viewer._widget.axes
                self._set_pos_and_margin(axes, *pos)
                viewer._widget.figure.canvas.draw()
            self._viewer_axes_positions = []
        # Record current positions if axes are currently hidden and hide them
        else:
            self._hide_viewer_axes()

    def _toggle_toolbars(self):
        self._toolbars_visible = not self._toolbars_visible
        for viewer in self.cube_views:
            viewer._widget.toolbar.setVisible(self._toolbars_visible)

    def _toggle_hover_value(self):
        for viewer in self.cube_views:
            viewer._widget._is_tooltip_on = not viewer._widget._is_tooltip_on

    def _toggle_stats_display(self):
        self._stats_visible = not self._stats_visible
        for viewer in self.cube_views:
            viewer.set_stats_visible(self._stats_visible)

    def _open_dialog(self, name, widget):

        if name == 'Collapse Cube':
            ex = collapse_cube.CollapseCube(
                self._wavelength_controller.wavelengths,
                self._wavelength_controller.current_units,
                self._data, parent=self, allow_preview=True)

        if name == 'Spatial Smoothing':
            ex = smoothing.SelectSmoothing(self._data, parent=self, allow_preview=True)

        if name == 'Arithmetic Operations':
            dialog = ArithmeticEditorWidget(self.session.data_collection)
            dialog.exec_()

        if name == "Moment Maps":
            mm_gui = moment_maps.MomentMapsGUI(
                self._data, self.session.data_collection, parent=self)
            mm_gui.display()

        if name == 'Convert Flux Units':
            self._flux_unit_controller.converter(parent=self)

        if name == "Wavelength Units/Redshift":
            WavelengthUI(self._wavelength_controller, parent=self)

    def refresh_flux_units(self, message):
        # TODO: eventually specviz should be able to respond to a
        # FluxUnitsUpdateMessage on its own.
        comp = self.specviz._widget._options_widget.file_att
        if message.component_id == comp:
            specviz_unit = message.flux_units
            if specviz_unit is not None:
                dispatch.changed_units.emit(y=specviz_unit)

    def _toggle_all_coords_in_degrees(self):
        """
        Switch ra-dec b/w "Sexagesimal" and "Decimal Degrees"
        """
        menu = self.ra_dec_format_menu
        for action in menu.actions():
            if "Decimal Degrees" == action.text():
                coords_in_degrees = action.isChecked()
                break

        for view in self.cube_views:
            viewer = view.widget()
            if viewer._coords_in_degrees != coords_in_degrees:
                viewer.toggle_coords_in_degrees()

    @property
    def data_components(self):
        return self._data.main_components + self._data.derived_components

    @property
    def component_labels(self):
        return [str(cid) for cid in self.data_components]

    def refresh_viewer_combo_helpers(self):
        for i, helper in enumerate(self._viewer_combo_helpers):
            helper.refresh()

    @dispatch.register_listener("apply_operations")
    def apply_to_cube(self, stack):
        """
        Listen for messages from specviz about possible spectral analysis
        operations that may be applied to the entire cube.
        """

        # Retrieve the current cube data object
        operation_handler = SpectralOperationHandler(self._data,
                                                     stack=stack,
                                                     session=self.session,
                                                     parent=self)
        operation_handler.exec_()

    def remove_data_component(self, component_id):
        pass

    def _enable_option_buttons(self):
        for button in self._option_buttons:
            button.setEnabled(True)
        self.ui.sync_button.setEnabled(True)

    def _get_change_viewer_combo_func(self, combo, view_index):

        def _on_viewer_combo_change(dropdown_index):

            # This function gets called whenever one of the viewer combos gets
            # changed. The active combo is the one that comes from the parent
            # _get_change_viewer_combo_func function.

            # Find the relevant viewer
            viewer = self.cube_views[view_index].widget()

            # Get the label of the component and the component ID itself
            label = combo.currentText()
            component = combo.currentData()
            if isinstance(component, UserDataWrapper):
                component = component.data

            viewer.has_2d_data = component.parent[label].ndim == 2

            # If the user changed the current component, stop previewing
            # smoothing.
            if viewer.is_smoothing_preview_active:
                viewer.end_smoothing_preview()

            # Change the component label, title and unit shown in the viewer
            viewer.current_component_id = component
            viewer.cubeviz_unit = self._flux_unit_controller.get_component_unit(component,
                                                                                cubeviz_unit=True)
            viewer.update_component_unit_label(component)
            viewer.update_axes_title(title=str(label))

            # Change the viewer's reference data to be the data containing the
            # current component.
            viewer.state.reference_data = component.parent

            # The viewer may have multiple layers, for instance layers for
            # the main cube and for any overlay datasets, as well as subset
            # layers. We go through all the layers and make sure that for the
            # layer which corresponds to the current dataset, the correct
            # attribute is shown.
            for layer_artist in viewer.layers:
                layer_state = layer_artist.state
                if layer_state.layer is component.parent:

                    # We call _update_attribute here manually so that if this
                    # function gets called before _update_attribute, it gets
                    # called before we try and set the attribute below
                    # (_update_attribute basically updates the internal list
                    # of available attributes for the attribute combo)
                    layer_state._update_attribute()
                    layer_state.attribute = component

                    # We then also make sure that this layer artist is the
                    # one that is selected so that if the user uses e.g. the
                    # contrast tool, it will change the right layer
                    viewer._view.layer_list.select_artist(layer_artist)

            # If the combo corresponds to the currently active cube viewer,
            # either activate or deactivate the slice slider as appropriate.
            if self.cube_views[view_index] is self._active_cube:
                self._slice_controller.set_enabled(not viewer.has_2d_data)

            # If contours are being currently shown, we need to force a redraw
            if viewer.is_contour_active:
                viewer.draw_contour()

            viewer.update_component(component)

        return _on_viewer_combo_change

    def _enable_viewer_combo(self, viewer, data, index, selection_label):
        connect_combo_selection(self, selection_label, viewer.combo)
        helper = ComponentIDComboHelper(self, selection_label)
        helper.set_multiple_data([data])
        viewer.combo.setEnabled(True)
        viewer.combo.currentIndexChanged.connect(
            self._get_change_viewer_combo_func(viewer.combo, index))
        self._viewer_combo_helpers.append(helper)

    def _enable_all_viewer_combos(self, data):
        """
        Setup the dropdown boxes that correspond to each of the left, middle,
        and right views.  The combo boxes initially are set to have FLUX,
        Error, DQ but will be dynamic depending on the type of data available
        either from being loaded in or by being processed.

        :return:
        """
        view = self.cube_views[0].widget()
        self._enable_viewer_combo(view.parent(), data, 0, 'single_viewer_attribute')

        component = self.single_viewer_attribute
        view.current_component_id = component
        view.cubeviz_unit = self._flux_unit_controller.get_component_unit(component,
                                                                          cubeviz_unit=True)
        view.update_component_unit_label(component)
        view.update_axes_title(component.label)

        for i in range(1,4):
            view = self.cube_views[i].widget()
            selection_label = 'viewer{0}_attribute'.format(i)
            self._enable_viewer_combo(view.parent(), data, i, selection_label)

            component = getattr(self, selection_label)
            view.current_component_id = component
            view.cubeviz_unit = self._flux_unit_controller.get_component_unit(component,
                                                                              cubeviz_unit=True)
            view.update_component_unit_label(component)
            view.update_axes_title(component.label)

    def change_viewer_component(self, view_index, component_id, force=False):
        """
        Given a viewer at an index view_index, change combo
        selection to component at an index component_index.
        :param view_index: int: Viewer index
        :param component_id: ComponentID: Component ID in viewer combo
        :param force: bool: force change if component is already displayed.
        """

        combo = self.get_viewer_combo(view_index)

        try:
            if isinstance(component_id, str):
                component_index = _find_combo_text(combo, component_id)
            else:
                component_index = _find_combo_data(combo, component_id)
        except ValueError:
            component_index = -1

        if combo.currentIndex() == component_index and force:
            combo.currentIndexChanged.emit(component_index)
        else:
            combo.setCurrentIndex(component_index)

    def display_component(self, component_id):
        """
        Displays data with given component ID in the active cube viewer.
        """
        self.refresh_viewer_combo_helpers()
        if self._single_viewer_mode:
            self.change_viewer_component(0, component_id)
        else:
            self.change_viewer_component(1, component_id)

    def get_viewer_combo(self, view_index):
        """
        Get viewer combo for a given viewer index
        """
        return self.cube_views[view_index].combo

    def add_overlay(self, data, label, display_now=True):
        self._overlay_controller.add_overlay(data, label, display=display_now)
        self.display_component(label)

    def _set_data_coord_system(self, data):
        """
        Check if data coordinates are in
        RA-DEC first. Then set viewers to
        the default coordinate system.
        :param data: input data
        """
        is_ra_dec = isinstance(wcs_to_celestial_frame(data.coords.wcs),
                               BaseRADecFrame)
        self.ra_dec_format_menu.setDisabled(not is_ra_dec)
        if not is_ra_dec:
            return

        is_coords_in_degrees = False
        for view in self.cube_views:
            viewer = view.widget()
            viewer.init_ra_dec()
            is_coords_in_degrees = viewer._coords_in_degrees

        if is_coords_in_degrees:
            format_name = "Decimal Degrees"
        else:
            format_name = "Sexagesimal"

        menu = self.ra_dec_format_menu
        for action in menu.actions():
            if format_name == action.text():
                action.setChecked(True)
                break

    def add_data(self, data):
        """
        Called by a function outside the class in order to add data to cubeviz.

        :param data:
        :return:
        """
        self._data = data
        self.specviz._widget.add_data(data)
        self._flux_unit_controller.set_data(data)

        comp = self.specviz._widget._options_widget.file_att
        specviz_unit = self._flux_unit_controller.get_component_unit(comp)
        if specviz_unit is not None:
            dispatch.changed_units.emit(y=specviz_unit)

        for checkbox in self._synced_checkboxes:
            checkbox.setEnabled(True)

        self._has_data = True
        self._active_view = self.split_views[0]
        self._active_cube = self.split_views[0]
        self._last_active_view = self.single_view
        self._active_split_cube = self.split_views[0]

        # Store pointer to wcs and wavelength information
        wcs = self.session.data_collection.data[0].coords.wcs
        wavelengths = self.single_view._widget._data[0].coords.world_axis(
            self.single_view._widget._data[0], axis=0)

        self._enable_all_viewer_combos(data)

        # TODO: currently this way of accessing units is not flexible
        self._slice_controller.enable()
        self._wavelength_controller.enable(str(wcs.wcs.cunit[2]), wavelengths)

        self._enable_option_buttons()
        self._setup_syncing()

        for viewer in self.cube_views:
            viewer.slice_text.setText('slice: {:5}'.format(self.synced_index))

        self.subWindowActivated.emit(self._active_view)

        # Check if coord system is RA and DEC (ie not galactic etc..)
        self._set_data_coord_system(data)


    def eventFilter(self, obj, event):

        if event.type() == QtCore.QEvent.MouseButtonPress:

            if not (self.isVisible() and self.isActiveWindow()):
                return super(CubeVizLayout, self).eventFilter(obj, event)

            # Find global click position
            click_pos = event.globalPos()

            # If the click position is the same as the last one, we shouldn't
            # do anything.
            if click_pos != self._last_click:

                # Determine if the event falls inside any of the viewers
                for viewer in self.subWindowList():
                    relative_click_pos = viewer.mapFromGlobal(click_pos)
                    if viewer.rect().contains(relative_click_pos):
                        # We should only emit an event if the active subwindow
                        # has actually changed.
                        if viewer is not self.activeSubWindow():
                            self.subWindowActivated.emit(viewer)
                        break

                self._last_click = click_pos

        return super(CubeVizLayout, self).eventFilter(obj, event)

    def _toggle_image_mode(self, event=None):
        new_active_view = self._last_active_view
        self._last_active_view = self._active_view

        # Currently in single image, moving to split image
        if self._single_viewer_mode:
            self._active_cube = self._active_split_cube
            self._activate_split_image_mode(event)
            self._single_viewer_mode = False
            self.ui.button_toggle_image_mode.setText('Single Image Viewer')

            for view in self.split_views:
                if self.single_view._widget.synced:
                    if view._widget.synced:
                        view._widget.update_slice_index(self.single_view._widget.slice_index)
                view._widget.update()
        # Currently in split image, moving to single image
        else:
            self._active_split_cube = self._active_cube
            self._active_view = self.single_view
            self._active_cube = self.single_view
            self._activate_single_image_mode(event)
            self._single_viewer_mode = True
            self.ui.button_toggle_image_mode.setText('Split Image Viewer')
            self._active_view._widget.update()

        self.subWindowActivated.emit(new_active_view)

        # Update the slice index to reflect the state of the active cube
        self._slice_controller.update_index(self._active_cube._widget.slice_index)

    def _activate_single_image_mode(self, event=None):
        vsplitter = self.ui.vertical_splitter
        hsplitter = self.ui.horizontal_splitter
        vsizes = list(vsplitter.sizes())
        hsizes = list(hsplitter.sizes())
        vsizes = 0, max(10, vsizes[0] + vsizes[1])
        hsizes = max(10, sum(hsizes) * 0.4), max(10, sum(hsizes) * 0.6)
        vsplitter.setSizes(vsizes)
        hsplitter.setSizes(hsizes)

    def _activate_split_image_mode(self, event=None):
        vsplitter = self.ui.vertical_splitter
        hsplitter = self.ui.horizontal_splitter
        vsizes = list(vsplitter.sizes())
        hsizes = list(hsplitter.sizes())
        vsizes = max(10, sum(vsizes) / 2), max(10, sum(vsizes) / 2)

        # TODO:  Might be a bug here, should the hsizes be based on vsizes? If so, not sure we need to calculate
        # TODO:  the hsizes above.
        hsizes = 0, max(10, vsizes[0] + vsizes[1])
        vsplitter.setSizes(vsizes)
        hsplitter.setSizes(hsizes)

    def _update_active_view(self, view):
        if self._has_data:
            self._active_view = view
            if isinstance(view._widget, CubevizImageViewer):
                self._active_cube = view
                index = self._active_cube._widget.slice_index
                if view._widget.has_2d_data:
                    self._slice_controller.set_enabled(False)
                else:
                    self._slice_controller.set_enabled(True)
                    self._slice_controller.update_index(index)

    def activeSubWindow(self):
        return self._active_view

    def subWindowList(self):
        return self.cube_views + [self.specviz]

    def _setup_syncing(self):
        for attribute in ['x_min', 'x_max', 'y_min', 'y_max']:
            # TODO: this will need to be generalized if we want to support an
            # arbitrary number of viewers.
            sync1 = keep_in_sync(self.split_views[0]._widget.state, attribute,
                                 self.split_views[1]._widget.state, attribute)
            sync2 = keep_in_sync(self.split_views[1]._widget.state, attribute,
                                 self.split_views[2]._widget.state, attribute)
            self.sync[attribute] = sync1, sync2
        self._on_sync_click()

    def _on_sync_click(self, event=None):
        index = self._active_cube._widget.slice_index
        for view in self.cube_views:
            view._widget.synced = True
            if view != self._active_cube:
                view._widget.update_slice_index(index)
        self._slice_controller.update_index(index)

    def start_smoothing_preview(self, preview_function, component_id, preview_title=None):
        """
        Starts smoothing preview. This function preforms the following steps
        1) SelectSmoothing passes parameters.
        2) The left and single viewers' combo box is set to component_id
        3) The set_smoothing_preview is called to setup on the fly smoothing
        :param preview_function: function: Single-slice smoothing function
        :param component_id: int: Which component to preview
        :param preview_title: str: Title displayed when previewing
        """
        # For single and first viewer:
        self._original_components = {}
        for view_index in [0, 1]:
            combo = self.get_viewer_combo(view_index)
            self._original_components[view_index] = combo.currentData()
            view = self.cube_views[view_index].widget()
            self.change_viewer_component(view_index, component_id, force=True)
            view.set_smoothing_preview(preview_function, preview_title)

    def end_smoothing_preview(self):
        """
        End preview and change viewer combo index to the first component.
        """
        for view_index in [0, 1]:
            view = self.cube_views[view_index].widget()
            view.end_smoothing_preview()
            if view_index in self._original_components:
                component_id = self._original_components[view_index]
                self.change_viewer_component(view_index, component_id, force=True)
        self._original_components = {}

    def showEvent(self, event):
        super(CubeVizLayout, self).showEvent(event)
        # Make split image mode the default layout
        self._activate_split_image_mode()
        self._update_active_view(self.split_views[0])

    def change_slice_index(self, amount):
        self._slice_controller.change_slider_value(amount)

    def get_wavelengths(self):
        return self._wavelength_controller.wavelengths

    def get_wavelengths_units(self):
        return self._wavelength_controller.current_units

    def get_wavelength(self, index=None):
        if index is None:
            index = self.synced_index
        elif index > len(self.get_wavelengths()):
            return None
        wave = self.get_wavelengths()[index]
        units = self.get_wavelengths_units()
        return wave * units



from glue.core.message import Message


def glue_subscribe(message):
    def decorator(function):
        self = args[0]
        self._hub.subscribe(self, message, handler=function)

    return decorator


class SliceIndexUpdateMessage(Message):

    # TODO: this should also track the originating layout tab so that if we
    # support multiple CubeViz tabs, the messages from different tabs don't
    # conflict
    def __init__(self, sender, index, data_set, slider_down=False, tag=None):
        super(SliceIndexUpdateMessage, self).__init__(sender, tag=tag)
        self.index = index
        self.data_set = data_set
        self.slider_down = slider_down


class WavelengthUpdateMessage(Message):

    def __init__(self, sender, wavelengths, tag=None):
        super(WavelengthUpdateMessage, self).__init__(sender, tag=tag)
        self.wavelengths = wavelengths


class WavelengthUnitUpdateMessage(Message):

    def __init__(self, sender, units, tag=None):
        super(WavelengthUnitUpdateMessage, self).__init__(sender, tag=tag)
        self.units = units


class RedshiftUpdateMessage(Message):

    def __init__(self, sender, redshift, label='', tag=None):
        super(RedshiftUpdateMessage, self).__init__(sender, tag=tag)
        self.redshift = redshift
        self.label = label


class FluxUnitsUpdateMessage(Message):

    def __init__(self, sender, flux_units, component_id, tag=None):
        super(FluxUnitsUpdateMessage, self).__init__(sender, tag=tag)
        self.flux_units = flux_units
        self.component_id = component_id



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import os

import pytest
import numpy as np

from cubeviz.tools.moment_maps import MomentMapsGUI

from .helpers import (toggle_viewer, select_viewer, left_click,
                      left_button_press, right_button_press, enter_slice_text,
                      assert_all_viewer_indices, assert_slice_text)


DATA_LABELS = ['018.DATA', '018.NOISE']


@pytest.fixture(scope='module')
def moment_map(cubeviz_layout):
    cl = cubeviz_layout
    mm_gui = MomentMapsGUI(cl._data, cl.session.data_collection, parent=cl)
    mm_gui.do_calculation(4, DATA_LABELS[0])

    return mm_gui.label

def test_starting_state(cubeviz_layout):
    assert cubeviz_layout.isVisible() == True

    assert cubeviz_layout._single_viewer_mode == False
    assert cubeviz_layout._active_view is cubeviz_layout.split_views[0]
    assert cubeviz_layout._active_cube is cubeviz_layout.split_views[0]

    for viewer in cubeviz_layout.cube_views:
        assert viewer._widget.synced == True

def assert_active_view_and_cube(layout, viewer):
    assert layout._active_view is viewer
    assert layout._active_cube is viewer

def test_active_viewer(qtbot, cubeviz_layout):
    select_viewer(qtbot, cubeviz_layout.split_views[1])
    assert_active_view_and_cube(cubeviz_layout, cubeviz_layout.split_views[1])

    select_viewer(qtbot, cubeviz_layout.split_views[2])
    assert_active_view_and_cube(cubeviz_layout, cubeviz_layout.split_views[2])

    select_viewer(qtbot, cubeviz_layout.specviz)
    assert cubeviz_layout._active_view is cubeviz_layout.specviz
    # Selecting the specviz viewer should not affect the last active cube
    assert cubeviz_layout._active_cube is cubeviz_layout.split_views[2]

    select_viewer(qtbot, cubeviz_layout.split_views[0])
    assert_active_view_and_cube(cubeviz_layout, cubeviz_layout.split_views[0])

def test_toggle_viewer_mode(qtbot, cubeviz_layout):
    toggle_viewer(qtbot, cubeviz_layout)
    assert cubeviz_layout._single_viewer_mode == True
    assert_active_view_and_cube(cubeviz_layout, cubeviz_layout.single_view)

    toggle_viewer(qtbot, cubeviz_layout)
    assert cubeviz_layout._single_viewer_mode == False
    assert_active_view_and_cube(cubeviz_layout, cubeviz_layout.split_views[0])

def test_remember_active_viewer(qtbot, cubeviz_layout):
    """Make sure that the active viewer in the current layout is remembered"""

    # Change active viewer in the split mode viewer
    select_viewer(qtbot, cubeviz_layout.split_views[2])
    assert_active_view_and_cube(cubeviz_layout, cubeviz_layout.split_views[2])

    # Change to the single mode viewer
    toggle_viewer(qtbot, cubeviz_layout)
    assert_active_view_and_cube(cubeviz_layout, cubeviz_layout.single_view)
    # Change active viewer in the single mode viewer
    select_viewer(qtbot, cubeviz_layout.specviz)
    assert cubeviz_layout._active_view == cubeviz_layout.specviz
    assert cubeviz_layout._active_cube == cubeviz_layout.single_view

    # Change back to the split mode viewer
    toggle_viewer(qtbot, cubeviz_layout)
    assert_active_view_and_cube(cubeviz_layout, cubeviz_layout.split_views[2])

    # Change back to the single mode viewer
    toggle_viewer(qtbot, cubeviz_layout)
    assert cubeviz_layout._active_view == cubeviz_layout.specviz
    assert cubeviz_layout._active_cube == cubeviz_layout.single_view

@pytest.mark.parametrize('viewer_index', [0, 1, 2, 3])
def test_sync_checkboxes(qtbot, cubeviz_layout, viewer_index):
    """This test simply makes sure that the checkbox changes the synced state
    of the corresponding viewer. It does not test the effects of syncing on the
    viewer indices.
    """
    # When testing the single viewer checkbox, first toggle to single viewer mode
    if viewer_index == 0:
        toggle_viewer(qtbot, cubeviz_layout)

    checkbox = cubeviz_layout._synced_checkboxes[viewer_index]
    viewer = cubeviz_layout.cube_views[viewer_index]

    left_click(qtbot, checkbox)
    assert viewer._widget.synced == False

    left_click(qtbot, checkbox)
    assert viewer._widget.synced == True

def check_data_component(layout, combo, index, widget):
    combo.setCurrentIndex(index)
    current_label = DATA_LABELS[index]
    assert combo.currentText() == current_label
    np.testing.assert_allclose(
        widget.layers[0].state.get_sliced_data(),
        layout._data[current_label][layout.synced_index])

def setup_combo_and_index(qtbot, layout, index):
    combo = layout.get_viewer_combo(index)
    if index == 0:
        toggle_viewer(qtbot, layout)
        synced_index = 0
    else:
        synced_index = index - 1

    return combo, synced_index

@pytest.mark.parametrize('viewer_index', [0, 1, 2, 3])
def test_viewer_dropdowns(qtbot, cubeviz_layout, viewer_index):
    combo, synced_index = setup_combo_and_index(
                                qtbot, cubeviz_layout, viewer_index)
    combo = cubeviz_layout.get_viewer_combo(viewer_index)
    if viewer_index == 0:
        toggle_viewer(qtbot, cubeviz_layout)
        synced_index = 0
    else:
        synced_index = min(viewer_index - 1, 1) # only two datasets

    widget = cubeviz_layout.cube_views[viewer_index]._widget

    # Make sure there are only two data components currently (dataset has two)
    assert combo.count() == 2
    # Make sure starting index is set appropriately
    assert combo.currentIndex() == synced_index

    for i in range(2):
        synced_index = (synced_index + 1) % 2
        check_data_component(cubeviz_layout, combo, synced_index, widget)

def test_add_data_component(qtbot, cubeviz_layout):
    new_label = 'QuirkyLabel'
    new_data = np.random.random(cubeviz_layout._data.shape)
    cubeviz_layout._data.add_component(new_data, new_label)

    for viewer_index in range(4):
        combo, synced_index = setup_combo_and_index(
                                    qtbot, cubeviz_layout, viewer_index)
        widget = cubeviz_layout.cube_views[viewer_index]._widget

        # Make sure the new index is there
        assert combo.count() == 3
        # Make sure the index hasn't changed (this might behave differently in the future)
        #assert combo.currentIndex() == synced_index

        # Make sure none of the original components have changed
        for i in range(2):
            check_data_component(cubeviz_layout, combo, i, widget)

        # Make sure the new data is displayed when selected
        combo.setCurrentIndex(2)
        assert combo.currentText() == new_label
        np.testing.assert_allclose(
            widget.layers[0].state.get_sliced_data(),
            cubeviz_layout._data[new_label][cubeviz_layout.synced_index])

        # Toggle back to split viewer mode if necessary
        if viewer_index == 0:
            toggle_viewer(qtbot, cubeviz_layout)

def test_key_shortcuts(qtbot, cubeviz_layout):
    slice_val = cubeviz_layout._slice_controller._slice_slider.value()
    left_button_press(qtbot, cubeviz_layout)
    assert cubeviz_layout._slice_controller._slice_slider.value() == slice_val - 1
    right_button_press(qtbot, cubeviz_layout)
    assert cubeviz_layout._slice_controller._slice_slider.value() == slice_val
    right_button_press(qtbot, cubeviz_layout)
    assert cubeviz_layout._slice_controller._slice_slider.value() == slice_val + 1

def assert_slider_enabled(cubeviz_layout, enabled):
    assert cubeviz_layout._slice_controller._slice_slider.isEnabled() == enabled
    assert cubeviz_layout._slice_controller._slice_textbox.isEnabled() == enabled
    assert cubeviz_layout._slice_controller._wavelength_textbox.isEnabled() == enabled


def test_2d_mouseover(cubeviz_layout, moment_map):

    combo = cubeviz_layout.get_viewer_combo(1)
    assert combo.currentText() == moment_map

    viewer = cubeviz_layout.split_views[0]._widget

    class FakeMPLEvent:
        xdata = 10
        ydata = 10
        inaxes = True

    viewer.mouse_move(FakeMPLEvent())

    value = "{:.3e}".format(cubeviz_layout._data.container_2d[moment_map][10][10])
    string = value + ' ' + str((10, 10, viewer.slice_index))

    assert viewer.mouse_value.startswith(value)
    assert viewer.coord_label.text().startswith(string)

@pytest.mark.parametrize('while_active', [True, False])
def test_2d_data_components(qtbot, cubeviz_layout, moment_map, while_active):
    """This test ensures that the slider behaves as expected when 2D data
    components are present. It tests updating the viewer containing 2D data
    both while it is the active viewer and while it is not the active
    viewer."""

    assert moment_map == DATA_LABELS[0] + '-moment-4'

    combo, _ = setup_combo_and_index(qtbot, cubeviz_layout, 1)

    # For the first test, the active viewer should now display a 2D moment map,
    # so no action is required.
    if not while_active:
        select_viewer(qtbot, cubeviz_layout.split_views[0])
        combo.setCurrentIndex(combo.findText(moment_map))

    assert_slider_enabled(cubeviz_layout, False)
    assert cubeviz_layout.split_views[0]._widget.has_2d_data == True
    assert cubeviz_layout.split_views[0]._widget.synced == False

    # Change the active viewer and make sure the slider is re-enabled
    select_viewer(qtbot, cubeviz_layout.split_views[1])
    assert_slider_enabled(cubeviz_layout, True)

    enter_slice_text(qtbot, cubeviz_layout, 1234)

    # Change back to the left viewer currently displaying 2D data
    if while_active:
        select_viewer(qtbot, cubeviz_layout.split_views[0])
        assert cubeviz_layout.split_views[0]._widget.has_2d_data == True
        assert cubeviz_layout.split_views[0]._widget.synced == False

    # Return to displaying 3D data component
    combo.setCurrentIndex(combo.findText(DATA_LABELS[0]))
    assert_all_viewer_indices(cubeviz_layout, 1234)
    assert cubeviz_layout.split_views[0]._widget.has_2d_data == False
    assert cubeviz_layout.split_views[0]._widget.synced == True

    if not while_active:
        select_viewer(qtbot, cubeviz_layout.split_views[0])
        assert_slider_enabled(cubeviz_layout, True)

    assert_slider_enabled(cubeviz_layout, True)
    assert_slice_text(cubeviz_layout, 1234)


def test_fast_slider_indexing(cubeviz_layout):
    lay = cubeviz_layout

    # Check if layer state is pointing to
    # correct viewer state
    view = lay.cube_views[1]._widget
    layer = view.layers[0]
    assert layer.state.viewer_state is view.state

    # Check if images are different when slice
    # index override is used.

    # Set viewer index to 0 and get slice data
    view.update_slice_index(0)
    arr1 = layer.state.get_sliced_data()

    # Change slice using slice index override
    # and get data.
    view.state.slice_index_override = 700
    arr2 = layer.state.get_sliced_data()

    # Check if the two images are the same: we expect that they are not
    assert np.any(arr1 != arr2)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
This is an Astropy affiliated package.
"""

import astropy.units as units
from .flux_equivalences import CustomFluxEquivalences
# We override the units.equivalencies.spectral_density function with
# CustomFluxEquivalences before the program starts. We expect all libraries
# to access CustomFluxEquivalences when calling for units.equivalencies.spectral_density
units.equivalencies.spectral_density = CustomFluxEquivalences(units.equivalencies.spectral_density)
units.spectral_density = units.equivalencies.spectral_density

# Affiliated packages may add whatever they like to this file, but
# should keep this content at the top.
# ----------------------------------------------------------------------------
from ._internal_init import *
# ----------------------------------------------------------------------------

if _CUBEVIZ_SETUP_ is False:
    from . import keyboard_shortcuts



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from os.path import basename, splitext
import yaml
import os
import sys
import glob
import logging

from glue.core import Data, Subset
from glue.core.coordinates import coordinates_from_header
from glue.config import data_factory, data_exporter
from glue.core.data_exporters.gridded_fits import fits_writer
from astropy.io import fits
import numpy as np

from cubeviz.data_factories.ifucube import IFUCube
from ..listener import CUBEVIZ_LAYOUT

from glue.utils.qt import load_ui

from qtpy.QtWidgets import (
    QDialog, QApplication, QPushButton)

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger('cubeviz_data_configuration')
logger.setLevel(logging.INFO)

DEFAULT_DATA_CONFIGS = os.path.join(os.path.dirname(__file__), 'configurations')
CUBEVIZ_DATA_CONFIGS = 'CUBEVIZ_DATA_CONFIGS'


class DataConfiguration:
    """
    This class is used to parse a YAML configuration file.

    """

    def __init__(self, config_file, check_ifu_valid=True):
        """
        Given the configuration file, save it and grab the name and priority
        :param config_file:
        """
        self._config_file = config_file
        self._check_ifu_valid = check_ifu_valid

        # If we are testing, then don't display the popup
        if self._check_ifu_valid:
            self.popup_ui = load_ui('ifucube_popup.ui', None, directory=os.path.dirname(__file__))

        with open(self._config_file, 'r') as ymlfile:
            cfg = yaml.load(ymlfile)

            self._name = cfg['name']
            self._type = cfg['type']


            try:
                self._priority = int(cfg.get('priority', 0))
            except Exception:
                self._priority = 0

            self._configuration = cfg['match']

            self._data = cfg.get('data', None)

            if 'flux_unit_replacements' in cfg:
                self.flux_unit_replacements = cfg['flux_unit_replacements']
            else:
                self.flux_unit_replacements = {}

    @property
    def name(self):
        return self._name

    @property
    def type(self):
        return self._type

    def get_units(self, header):
        """
        Extract BUNIT from header.
        BUNIT contains the flux units
            KEYWORD:   BUNIT
            REFERENCE: FITS Standard
            STATUS:    reserved
            HDU:       image
            VALUE:     string
            COMMENT:   physical units of the array values
            DEFINITION: The value field shall contain a character string,
            describing the physical units in which the quantities in the array,
            after application of BSCALE and BZERO, are expressed.   The units of
            all FITS header keyword values, with the exception of measurements of
            angles, should conform with the recommendations in the IAU Style
            Manual. For angular measurements given as floating point values and
            specified with reserved keywords, degrees are the recommended units
            (with the units, if specified, given as 'deg').
        For more info on header keys:
        https://heasarc.gsfc.nasa.gov/docs/fcg/standard_dict.html
        :param header: header
        :return: str: Shortened unit
        """
        units = str(header['BUNIT'])

        # Unit label shorten depending on data type here
        for key in self.flux_unit_replacements:
            if key in units:
                units = units.replace(key, self.flux_unit_replacements[key])
        return units

    def _reject_button_click(self):
        """
        Closes entire program if clicked
        :return:
        """
        self.popup_ui.close()
        sys.exit()

    def _accept_button_click(self):
        """
        Continues as usual
        :return:
        """
        self.popup_ui.accept()

    def load_data(self, data_filenames):
        """
        Load the data based on the extensions defined in the matching YAML file.  THen
        create the datacube and return it.

        :param data_filename:
        :return:
        """

        label = None
        data = None

        ifucube = IFUCube()

        for data_filename in data_filenames.split(','):

            hdulist = ifucube.open(data_filename, fix=self._check_ifu_valid)

            # Good in this case means the file has 3D data and can be loaded by SpectralCube.read
            if self._check_ifu_valid and not ifucube.get_good():
                # Popup takes precedence and accepting continues operation and canceling closes the program
                self.popup_ui.ifucube_log.setText(ifucube.get_log_output())
                self.popup_ui.setModal(True)
                self.popup_ui.show()

                self.popup_ui.button_accept.clicked.connect(self._accept_button_click)
                self.popup_ui.button_cancel.clicked.connect(self._reject_button_click)

            if not label:
                label = "{}: {}".format(self._name, splitext(basename(data_filename))[0])
                data = Data(label=label)

                # this attribute is used to indicate to the cubeviz layout that
                # this is a cubeviz-specific data component.
                data.meta[CUBEVIZ_LAYOUT] = self._name

            data_coords_set = False
            for ii, hdu in enumerate(hdulist):
                if 'NAXIS' in hdu.header and hdu.header['NAXIS'] == 3:

                    # Set the coords based on the first 3D HDU
                    if not data_coords_set:
                        data.coords = coordinates_from_header(hdu.header)
                        data_coords_set = True

                    component_name = str(ii)
                    if 'EXTNAME' in hdu.header:
                        component_name = hdu.header['EXTNAME']

                        # The data must be floating point as spectralcube is expecting floating point data
                        data.add_component(component=hdu.data.astype(np.float), label=component_name)

                        if 'BUNIT' in hdu.header:
                            c = data.get_component(component_name)
                            c.units = self.get_units(hdu.header)
                    else:
                        # Creates a unique component name
                        component_name = str(ii)
                        data.add_component(component=hdu.data.astype(np.float), label=component_name)

            # For the purposes of exporting, we keep a reference to the original HDUList object
            data._cubeviz_hdulist = hdulist

        return data

    def matches(self, filename):
        """
        Main call to which we pass in the file to see if it matches based
        on the criteria in the config file.

        :param filename:
        :return:
        """
        # Check the "first filename in the list" which might be the "only filename" in the list.
        filename = filename.split(',')[0]
        self._fits = fits.open(filename)

        # Now call the internal processing.
        matches = self._process('all', self._configuration['all'])

        if matches:
            logger.debug('{} matches {}'.format(self._config_file, filename))
        else:
            logger.debug('{} does not match {}'.format(self._config_file, filename))

        return matches

    def _process(self, key, conditional):
        """
        Internal processing. This will get called numerous times recursively.

        :param key: The type of check we want to do
        :param conditional: The thing we are checking
        :return:
        """

        if 'all' == key:
            return self._all(conditional)
        elif 'any' == key:
            return self._any(conditional)
        elif 'equal' == key:
            return self._equal(conditional)
        elif 'startswith' == key:
            return self._startswith(conditional)
        elif 'extension_names' == key:
            return self._extension_names(conditional)
        elif 'has_data' == key:
            return self._has_data()

    #
    # Branch processing
    #

    def _all(self, conditionals):
        """
        All conditions must be met

        :param conditionals:
        :return:
        """
        logger.debug('\tall: {}'.format(conditionals))
        for key, conditional in conditionals.items():
            ret = self._process(key, conditional)
            if not ret:
                return False
        return True

    def _any(self, conditionals):
        """
        Any condition can be met

        :param conditionals:
        :return:
        """

        logger.debug('\tany: {}'.format(conditionals))
        for key, conditional in conditionals.items():
            # process conditional
            ret = self._process(key, conditional)
            if ret:
                return ret

        return False

    #
    # Leaf processing
    #

    def _has_data(self):
        """
        Used for miscellaneous cubes that we have not planned for by just making sure data in the cube exists. This
        is currently only in use in the default.yaml file
        :return:
        """
        for hdu in self._fits:
            if hasattr(hdu, 'data') and hdu.data is not None and hasattr(hdu.data, 'shape') and len(hdu.data.shape) == 3:
                return True
        return False

    def _equal(self, value):
        """
        The value at the header_key must equal the value

        :param value:
        :return:
        """
        logger.debug('\tequality: {} = {} ?'.format(self._fits[0].header.get(value['header_key'], False), value['value']))
        return self._fits[0].header.get(value['header_key'], False) == value['value']

    def _startswith(self, value):
        """
        The value at the header_key must start with the value

        :param value:
        :return:
        """
        logger.debug('\tstartswith: {} starswith {} ?'.format(self._fits[0].header.get(value['header_key'], False), value['value']))
        return self._fits[0].header.get(value['header_key'], '').startswith(value['value'])

    def _extension_names(self, value):
        """
        All extensions must exist in the file

        :param value:
        :return:
        """
        logger.debug('\tcontains extension: {} in {} ?'.format(value,
                                                               [x.header['EXTNAME'] for x in self._fits if 'EXTNAME' in x.header]))

        if isinstance(value, str):
            return value in self._fits
        else:
            return all([v in self._fits for v in value])

    def summarize(self):
        """
        High level summarize function for the YAML configuration file.
        """

        print('# {}'.format(self._name))
        print('Description: {}\n'.format(self._type))
        print('Filename: {}\n'.format(self._config_file))
        print('\n')

        self._summarize(self._configuration)

    def _summarize(self, d, level=1):
        """
        Summarize function that will either print the lear values or will go
        one level deeper.
        """

        # Leaf value, so just print it out.
        if 'header_key' in d.keys():
            print("""{}* header['{}']   '{}'""".format('  '*level, d['header_key'], d['value']))
            print('\n')


        # Check each key of the dictionary and go deeper if needed.
        else:
            for k, v in d.items():

                func = eval('self._{}'.format(k))
                print('{}* {}:'.format('  '*level, self._get_func_docstring(func)))

                if isinstance(v, dict):
                    self._summarize(v, level+1)
                elif v:
                    print('{}{}'.format('  '*level, v))

    def _get_func_docstring(self, func):
        """
        Given the string representation of one of the functions, get the
        docstring and return it to be used in the markup.
        """
        ds = func.__doc__
        for dsl in ds.split('\n'):
            if len(dsl) > 0:
                return dsl.strip()

        return ''


class DataFactoryConfiguration:
    """
    This class takes in lists of files or directories that are or contain data configuration YAML files. These
    can come from:
       1. the default location
       2. from the command line "--data-configs <file-or-directory>"
       3. from the environment variable CUBEVIZ_DATA_CONFIGS=<files-or-directories>
    """
    def __init__(self, in_configs=[], show_only=False, remove_defaults=False, check_ifu_valid=True):
        """
        The IFC takes either a directory (that contains YAML files), a list of directories (each of which contain
        YAML files) or a list of YAML files.  Each YAML file defines requirements

        :param in_configs: Directory, list of directories, or list of files.
        """

        # Remove all pre-defined data configuration loaders in Glue. Then, if a user tries to open an IFU FITS
        # file that is not known to us a popup will come up saying cubeviz does not recognize the data format.
        if remove_defaults:
            data_factory._members = []

        if show_only:
            logger.setLevel(logging.DEBUG)

        self._config_files = []

        # Get all the available YAML data configuration files based on the command line.
        logger.debug('YAML data configuration fiels from command line: {}'.format(in_configs))
        self._config_files.extend(self._find_yaml_files(in_configs))

        # Get all the available YAML data configuration files based on the environment variable.
        if CUBEVIZ_DATA_CONFIGS in os.environ:
            logger.debug('YAML data configuration fiels from environment variable: {}'.format(os.environ[CUBEVIZ_DATA_CONFIGS]))
            self._config_files.extend(self._find_yaml_files(os.environ[CUBEVIZ_DATA_CONFIGS]))

        # Get all the available YAML data configuration files based on the default directory
        logger.debug(
            'YAML data configuration fiels from the default directory: {}'.format(DEFAULT_DATA_CONFIGS))
        self._config_files.extend(self._find_yaml_files(DEFAULT_DATA_CONFIGS))

        logger.debug(
            'YAML data configuration files: {}'.format('\n'.join(self._config_files)))

        for config_file in self._config_files:

            # Load the YAML file and get the name, priority and create the data factory wrapper
            with open(config_file, 'r') as yamlfile:
                cfg = yaml.load(yamlfile)

            name = cfg['name']

            try:
                priority = int(cfg.get('priority', 0))
            except:
                priority = 0

            # The code below instantiates a data configuration object based on the config file and is
            # therefore dependent on the type of data file.  The data configuration object defines two functions
            # 'matches' and 'load_data' that are used.  We needed a way to call Glue's data_factory and be able
            # to pass in functions that have state information.
            dc = DataConfiguration(config_file, check_ifu_valid=check_ifu_valid)
            wrapper = data_factory(name, dc.matches, priority=priority)
            wrapper(dc.load_data)



    def _find_yaml_files(self, files_or_directories):
        """
        Given the files_or_directories, create a list of all relevant YAML files.

        :param files_or_directories:
        :return:
        """
        config_files = []

        # If the thing passed in was a string then we'll split on colon. If there is only one
        # directory then it will create a list anyway.
        if isinstance(files_or_directories, str):
            files_or_directories = files_or_directories.split(':')

        for x in files_or_directories:
            if os.path.exists(x):
                # file, just append
                if os.path.isfile(x):
                    config_files.append(x)
                # directory, find all yaml under it.
                else:
                    files  = glob.glob(os.path.join(x, '*.yaml'))
                    config_files.extend(files)

        return config_files


    def summarize(self):
        """
        Function to print out a Markup representation of each configuration file.
        """
        for config_file in self._find_yaml_files(DEFAULT_DATA_CONFIGS):
            dc = DataConfiguration(config_file)
            print(dc.summarize())


@data_exporter('CubeViz FITS exporter', extension=['fits', 'fit'])
def cubeviz_fits_exporter(filename, data, components=None):

    if isinstance(data, Subset):
        raise NotImplementedError("Can't export subsets yet")

    if not hasattr(data, '_cubeviz_hdulist'):
        return fits_writer(filename, data, components=components)

    if components is None:
        components = data.visible_components

    hdulist = fits.HDUList(data._cubeviz_hdulist.copy())

    component_labels = [cid.label for cid in components]

    # Remove any HDUs with data that don't have a matching component
    for hdu in hdulist[::-1]:
        if hdu.data is not None and hdu.name not in component_labels:
            hdulist.remove(hdu)

    # Add any other components
    for cid in components:

        if cid.label in hdulist:
            continue

        comp = data.get_component(cid)

        if comp.categorical:
            raise NotImplementedError()

        hdu = fits.ImageHDU(comp.data, data.coords.wcs.to_header(), name=cid.label)
        hdulist.append(hdu)

    try:
        hdulist.writeto(filename, overwrite=True)
    except TypeError:
        hdulist.writeto(filename, clobber=True)



from __future__ import absolute_import, division, print_function

import re
import os
from collections import OrderedDict
import numpy as np

from qtpy.QtCore import Qt
from qtpy import QtGui
from qtpy.QtWidgets import (QDialog, QApplication, QPushButton, QLabel, QWidget,
                            QHBoxLayout, QVBoxLayout, QLineEdit, QComboBox)
from glue.utils.qt import load_ui

from astropy.stats import sigma_clip

from .common import add_to_2d_container, show_error_message

import logging
logging.basicConfig(format='%(levelname)-6s: %(name)-10s %(asctime)-15s  %(message)s')
log = logging.getLogger("CollapseCube")
log.setLevel(logging.WARNING)

# The operations we understand
operations = OrderedDict([
    ('Sum', np.sum),
    ('Mean', np.mean),
    ('Median', np.median),
    ('Standard Deviation', np.std),
    ('Maximum', np.max),
    ('Minimum', np.min),
    ('Sum (ignore NaNs)', np.nansum),
    ('Mean (ignore NaNs)', np.nanmean),
    ('Median (ignore NaNs)', np.nanmedian),
    ('Standard Deviation (ignore NaNs)', np.nanstd),
    ('Maximum (ignore NaNs)', np.nanmax),
    ('Minimum (ignore NaNs)', np.nanmin)
])

class CollapseCube(QDialog):
    def __init__(self, wavelengths, wavelength_units, data, data_collection=[],
                 allow_preview=False, parent=None):

        super(CollapseCube,self).__init__(parent)

        self.setWindowTitle("Collapse Cube Along Spectral Axis")

        # Get the data_components (e.g., FLUX, DQ, ERROR etc)
        # Using list comprehension to keep the order of the component_ids
        self.data_components = [str(x).strip() for x in data.component_ids() if not x in data.coordinate_components]

        self.setWindowFlags(self.windowFlags() | Qt.Tool)
        self.title = "Cube Collapse"

        self.wavelengths = wavelengths
        self.wavelength_units = wavelength_units
        self.data = data
        self.data_collection = data_collection
        self.parent = parent

        self._general_description = "Collapse the data cube over the spectral range based on the mathematical operation.  The nearest index or wavelength will be chosen if a specified number is out of bounds"
        self._custom_description = "To use the spectral viewer to define a region to collapse over, cancel this, create an ROI and then select this Collapse Cube again."

        self._extra_message = ''

        self.currentAxes = None
        self.currentKernel = None

        self.createUI()

    def createUI(self):
        """
        Create the popup box with the calculation input area and buttons.

        :return:
        """
        # Load the widget from the UI file.
        self.ui = load_ui('collapse.ui', self,
                          directory=os.path.dirname(__file__))

        self.ui.data_combobox.addItems([str(x).strip() for x in self.data.component_ids()
                                      if not x in self.data.coordinate_components])
        self.ui.operation_combobox.addItems(operations.keys())

        # Fill spatial region combobox
        spatial_regions = ['Image'] + [x.label for x in self.data.subsets]
        self.ui.spatial_region_combobox.addItems(spatial_regions)

        # Get the Specviz regions and add them in to the Combo box
        for roi in self.parent.specviz._widget.roi_bounds:
            self.ui.region_combobox.addItem("Specviz ROI ({:.4e}, {:.4e})".format(roi[0], roi[1]))

        self.ui.region_combobox.addItems(["Custom (Wavelengths)", "Custom (Indices)"])
        self.ui.region_combobox.setMinimumWidth(200)
        self.ui.region_combobox.currentIndexChanged.connect(self._region_selection_change)

        self.ui.desc_label.setText(self._general_description)

        # Set defaults for wavelengths
        indthird = len(self.wavelengths) // 3
        self.ui.start_input.setText('{:.4e}'.format(self.wavelengths[indthird]))
        self.ui.end_input.setText('{:.4e}'.format(self.wavelengths[2*indthird]))

        # Hide the error box... for now.
        self.ui.error_label.setVisible(False)

        # Setup call back for the Advanced Sigma Checkbox
        self.ui.region_combobox.currentIndexChanged.connect(self._region_combobox_callback)

        # Setup call back for the Advanced Sigma Checkbox
        self.ui.sigma_combobox.currentIndexChanged.connect(self._sigma_combobox_callback)

        # Setup the call back for the buttons
        self.ui.calculate_button.clicked.connect(self.calculate_callback)
        self.ui.cancel_button.clicked.connect(self.cancel_callback)

        self.ui.show()

        # Fire the callback to set the default values for everything
        self._sigma_combobox_callback(0)
        self._region_selection_change(0)

    def _region_combobox_callback(self, index):
        log.debug('_region_combobox_callback with index {}'.format(index))

        combo_text = self.ui.region_combobox.currentText()

        if 'Custom (Wavelengths)' == combo_text:  # convert to custom wavelengths
            # try to convert the text boxes to wavelengths
            start = self.ui.start_input.text().strip()
            log.debug('    start = {}'.format(start))
            try:
                if float(start).is_integer():
                    start = int(start)
                    self.ui.start_input.setText('{:.4e}'.format(self.wavelengths[int(start)]))
            except:
                self.ui.start_input.setText('')

            # try to convert the text boxes to wavelengths
            end = self.ui.end_input.text().strip()
            log.debug('    end = {}'.format(end))
            try:
                if float(end).is_integer():
                    end = int(end)
                    self.ui.end_input.setText('{:.4e}'.format(self.wavelengths[int(end)]))
            except:
                self.ui.end_input.setText('')

        elif 'Custom (Indices)' == combo_text:  # convert to custom indices
            # try to convert the text boxes to indices
            start = self.ui.start_input.text().strip()
            log.debug('    start = {}'.format(start))
            try:
                ind = np.argsort(abs(float(start) - self.wavelengths))[0]
                self.ui.start_input.setText('{}'.format(ind))
            except:
                self.ui.start_input.setText('')

            # try to convert the text boxes to wavelengths
            end = self.ui.end_input.text().strip()
            log.debug('    end = {}'.format(end))
            try:
                ind = np.argsort(abs(float(end) - self.wavelengths))[0]
                self.ui.end_input.setText('{}'.format(ind))
            except:
                self.ui.end_input.setText('')

    def _sigma_combobox_callback(self, index):

        # Show / Hide simple sigma options
        self.ui.simple_sigma_label.setVisible(index == 1)
        self.ui.simple_sigma_description.setVisible(index == 1)
        self.ui.simple_sigma_input.setVisible(index == 1)

        # Show / Hide advanced sigma options
        self.ui.advanced_sigma_label.setVisible(index == 2)
        self.ui.advanced_sigma_description.setVisible(index == 2)
        self.ui.advanced_sigma_lower_label.setVisible(index == 2)
        self.ui.advanced_sigma_upper_label.setVisible(index == 2)
        self.ui.advanced_sigma_iters_label.setVisible(index == 2)

        self.ui.advanced_sigma_input.setVisible(index == 2)
        self.ui.advanced_sigma_lower_input.setVisible(index == 2)
        self.ui.advanced_sigma_upper_input.setVisible(index == 2)
        self.ui.advanced_sigma_iters_input.setVisible(index == 2)

    def _region_selection_change(self, index):
        """
        Callback for a change on the region selection combo box.

        :param newvalue:
        :return:
        """

        newvalue = self.region_combobox.currentText()
        indthird = len(self.wavelengths) // 3

        # First, let's see if this is one of the custom options
        if 'Custom' in newvalue and 'Wavelength' in newvalue:
            # Custom Wavelengths
            self.hide_start_end(False)
            self.ui.start_label.setText("Start Wavelength:")
            self.ui.end_label.setText("End Wavelength:")

            self.ui.start_example_label.setText('(e.g., {:.4e})'.format(self.wavelengths[indthird]))
            self.ui.end_example_label.setText('(e.g., {:.4e})'.format(self.wavelengths[2*indthird]))

        elif 'Custom' in newvalue and 'Indices' in newvalue:
            # Custom indices
            self.hide_start_end(False)
            self.ui.start_label.setText("Start Index:")
            self.ui.end_label.setText("End Index:")

            self.ui.start_example_label.setText('(e.g., {})'.format(indthird))
            self.ui.end_example_label.setText('(e.g., {})'.format(2*indthird))

        else:
            # Region defined in specviz
            self.hide_start_end(True)

            # We are going to store the start and end wavelengths in the text boxes even though
            # they are hidden. This way we can use the text boxes later as a hidden storage container.
            # TODO: Should probably save the ROIs so the start and end values are more accurate.
            regex = r"-?(?:0|[1-9]\d*)(?:\.\d*)?(?:[eE][+\-]?\d+)?"
            floating = re.findall(regex, newvalue)
            self.ui.start_input.setText(floating[0])
            self.ui.end_input.setText(floating[1])

        # Let's update the text on the widget
        if 'Custom' in newvalue:
            self.ui.desc_label.setText(self._general_description + '\n' + self._custom_description)
        else:
            self.ui.desc_label.setText(self._general_description)

    def hide_start_end(self, dohide):
        """
        Show or hide the start and end indices depending if the region
        is defined from the specviz plot OR if we are using custom limits.

        :param dohide:
        :return:
        """
        self.ui.start_label.setEnabled(not dohide)
        self.ui.start_example_label.setEnabled(not dohide)
        self.ui.start_input.setEnabled(not dohide)
        self.ui.end_label.setEnabled(not dohide)
        self.ui.end_example_label.setEnabled(not dohide)
        self.ui.end_input.setEnabled(not dohide)

    def _calculate_callback_wavelength_checks(self, start_wavelength, end_wavelength):

        log.debug('_calculate_callback_wavelength_checks with start {} and end {}'.format(
            start_wavelength, end_wavelength))

        if len(start_wavelength) == 0:
            start_wavelength = self.wavelengths[0]
            self.ui.start_input.setText('{:.4e}'.format(start_wavelength))
            self._extra_message += ' Start wavelength not set so used the first.'

        if len(end_wavelength) == 0:
            end_wavelength = self.wavelengths[-1]
            self.ui.end_input.setText('{:.4e}'.format(end_wavelength))
            self._extra_message += ' End wavelength not set so used the last.'

        try:
            start_wavelength = float(start_wavelength)
        except:
            self.ui.start_label.setStyleSheet("color: rgba(255, 0, 0, 128)")
            self.ui.error_label.setText('Start wavelength is not a floating point number.')
            self.ui.error_label.setVisible(True)

            return None, None

        try:
            end_wavelength = float(end_wavelength)
        except:
            self.ui.start_label.setStyleSheet("color: rgba(255, 0, 0, 128)")
            self.ui.error_label.setText('End wavelength is not a floating point number.')
            self.ui.error_label.setVisible(True)

            return None, None

        if start_wavelength < self.wavelengths[0]:
            self.ui.error_label.setText('Start wavelength is out of range, setting to end point.')
            self.ui.error_label.setVisible(True)
            self._extra_message += ' Start wavelength too low, using the first wavelength.'

            start_wavelength = self.wavelengths[0]

        if start_wavelength > self.wavelengths[-1]:
            self.ui.start_label.setStyleSheet("color: rgba(255, 0, 0, 128)")
            self.ui.error_label.setText('Start wavelength is out of range.')
            self.ui.error_label.setVisible(True)
            return None, None

        if end_wavelength < start_wavelength:
            self.ui.start_label.setStyleSheet("color: rgba(255, 0, 0, 128)")
            self.ui.end_label.setStyleSheet("color: rgba(255, 0, 0, 128)")
            self.ui.error_label.setText('Start wavelength must be less than the end wavelength.')
            self.ui.error_label.setVisible(True)
            return None, None

        if end_wavelength > self.wavelengths[-1]:
            self.ui.error_label.setText('End wavelength is out of range, setting to end point.')
            self.ui.error_label.setVisible(True)
            self._extra_message += ' End wavelength too high, using the last wavelength.'

            end_wavelength = self.wavelengths[-1]

        if end_wavelength < self.wavelengths[0]:
            self.ui.end_label.setStyleSheet("color: rgba(255, 0, 0, 128)")
            self.ui.error_label.setText('End wavelength is out of range.')
            self.ui.error_label.setVisible(True)

            end_wavelength = None

        start_index = np.argsort(abs(self.wavelengths - start_wavelength))[0]
        end_index = np.argsort(abs(self.wavelengths - end_wavelength))[0]

        log.debug('  returning with start_index {} and end_index {}'.format(
            start_index, end_index))

        return start_index, end_index

    def _calculate_callback_index_checks(self, start_index, end_index):

        log.debug('_calculate_callback_index_checks with start {} and end {}'.format(
            start_index, end_index))

        if len(start_index) == 0:
            start_index = 0
            self.ui.start_input.setText('{}'.format(start_index))
            self._extra_message += ' Start wavelength not set so used the first.'

        if len(end_index) == 0:
            end_index = len(self.wavelengths) - 1
            self.ui.end_input.setText('{}'.format(end_index))
            self._extra_message += ' End wavelength not set so used the last.'

        try:
            start_index = int(start_index)
        except:
            self.ui.start_label.setStyleSheet("color: rgba(255, 0, 0, 128)")
            self.ui.error_label.setText('Start index is not an integer.')
            self.ui.error_label.setVisible(True)

            start_index = None

        try:
            end_index = int(end_index)
        except:
            self.ui.start_label.setStyleSheet("color: rgba(255, 0, 0, 128)")
            self.ui.error_label.setText('End index is not an integer.')
            self.ui.error_label.setVisible(True)

            end_index = None

        if start_index < 0:
            self.ui.error_label.setText('Start index is out of range, setting to end point.')
            self.ui.error_label.setVisible(True)
            self._extra_message += ' Start wavelength too low, using the first wavelength.'

            start_index = 0

        if start_index > len(self.wavelengths):
            self.ui.start_label.setStyleSheet("color: rgba(255, 0, 0, 128)")
            self.ui.error_label.setText('Start index is out of range.')
            self.ui.error_label.setVisible(True)

            start_index = None

        if end_index < start_index:
            self.ui.end_label.setStyleSheet("color: rgba(255, 0, 0, 128)")
            self.ui.error_label.setText('Start index must be less than the end index.')
            self.ui.error_label.setVisible(True)
            return None, None

        if end_index > len(self.wavelengths):
            self.ui.error_label.setText('End index is out of range, setting to end point.')
            self.ui.error_label.setVisible(True)
            self._extra_message += ' End wavelength too high, using the last wavelength.'

            end_index = len(self.wavelengths)-1

        if end_index < 0:
            self.ui.end_label.setStyleSheet("color: rgba(255, 0, 0, 128)")
            self.ui.error_label.setText('End index is out of range.')
            self.ui.error_label.setVisible(True)

            end_index = None

        log.debug('  returning with start_index {} and end_index {}'.format(
            start_index, end_index))

        return start_index, end_index

    def _calculate_callback_simple_sigma_check(self):
        log.debug('_calculate_callback_simgple_sigma_check')

        simple_sigma = self.ui.simple_sigma_input.text().strip()
        log.debug('    simple_sigma {}'.format(simple_sigma))

        try:
            simple_sigma = float(simple_sigma)
        except:
            self.ui.simple_sigma_label.setStyleSheet("color: rgba(255, 0, 0, 128)")
            self.ui.error_label.setText('Sigma must be a floating point number.')
            self.ui.error_label.setVisible(True)
            return None

        if simple_sigma <= 0.0:
            self.ui.simple_sigma_label.setStyleSheet("color: rgba(255, 0, 0, 128)")
            self.ui.error_label.setText('Sigma must be a positive number.')
            self.ui.error_label.setVisible(True)
            return None

        return simple_sigma

    def _calculate_callback_advanced_sigma_check(self):
        sigma = self.ui.advanced_sigma_input.text().strip()
        sigma_lower = self.ui.advanced_sigma_lower_input.text().strip()
        sigma_upper = self.ui.advanced_sigma_upper_input.text().strip()
        sigma_iters = self.ui.advanced_sigma_iters_input.text().strip()

        if len(sigma) > 0:
            try:
                sigma = float(sigma)
            except:
                self.ui.advanced_sigma_label.setStyleSheet("color: rgba(255, 0, 0, 128)")
                self.ui.error_label.setText('Sigma must be a floating point number.')
                self.ui.error_label.setVisible(True)
                return None
        else:
            self.ui.advanced_sigma_label.setStyleSheet("color: rgba(255, 0, 0, 128)")
            self.ui.error_label.setText('Sigma must be a floating point number and not empty.')
            self.ui.error_label.setVisible(True)
            return None, None, None, None

        if len(sigma_lower) > 0:
            try:
                sigma_lower = float(sigma_lower)
            except:
                self.ui.advanced_sigma_lower_label.setStyleSheet("color: rgba(255, 0, 0, 128)")
                self.ui.error_label.setText('Lower sigma must be a floating point number.')
                self.ui.error_label.setVisible(True)
                return [None]*4
        else:
            sigma_lower = None

        if len(sigma_upper) > 0:
            try:
                sigma_upper = float(sigma_upper)
            except:
                self.ui.advanced_sigma_upper_label.setStyleSheet("color: rgba(255, 0, 0, 128)")
                self.ui.error_label.setText('Upper sigma must be a floating point number.')
                self.ui.error_label.setVisible(True)
                return [None]*4
        else:
            sigma_upper = None

        if len(sigma_iters) > 0:
            try:
                sigma_iters = int(sigma_iters)
            except:
                self.ui.advanced_sigma_iters_label.setStyleSheet("color: rgba(255, 0, 0, 128)")
                self.ui.error_label.setText('Iterations for sigma must be a floating point number.')
                self.ui.error_label.setVisible(True)
                return [None]*4
        else:
            sigma_iters = None

        if sigma_lower is not None and sigma_upper is not None and sigma_lower > sigma_upper:
            self.ui.advanced_sigma_lower_label.setStyleSheet("color: rgba(255, 0, 0, 128)")
            self.ui.advanced_sigma_upper_label.setStyleSheet("color: rgba(255, 0, 0, 128)")
            self.ui.error_label.setText('Lower sigma must be smaller than the upper sigma.')
            self.ui.error_label.setVisible(True)
            return [None]*4

        if sigma <= 0.0:
            self.ui.advanced_sigma_label.setStyleSheet("color: rgba(255, 0, 0, 128)")
            self.ui.error_label.setText('Sigma must be a positive number.')
            self.ui.error_label.setVisible(True)
            return [None]*4

        return sigma, sigma_lower, sigma_upper, sigma_iters

    def _calculate_collapse(self, data_name, operation, spatial_region, sigma_selection, sigma_parameter, start_index, end_index):

        start_wavelength = self.wavelengths[start_index]
        end_wavelength = self.wavelengths[end_index]

        label = '{}-collapse-{} ({:.4e}, {:.4e})'.format(data_name, operation,
                                                         start_wavelength,
                                                         end_wavelength)

        # Setup the input_data (and apply the spatial mask based on
        # the selection in the spatial_region_combobox
        input_data = self.data[data_name]
        log.debug('    spatial region is {}'.format(spatial_region))
        if not spatial_region == 'Image':
            subset = [x.to_mask() for x in self.data.subsets if x.label == spatial_region][0]
            input_data = input_data * subset

        # Apply sigma clipping
        if 'Simple' in sigma_selection:
            sigma = sigma_parameter

            if sigma is None:
                return

            input_data = sigma_clip(input_data, sigma=sigma, axis=0)
            label += ' sigma={}'.format(sigma)

        elif 'Advanced' in sigma_selection:
            sigma, sigma_lower, sigma_upper, sigma_iters = sigma_parameter
            log.debug('    returned from calculate_callback_advanced_sigma_check with sigma {}  sigma_lower {}  sigma_upper {}  sigma_iters {}'.format(
                sigma, sigma_lower, sigma_upper, sigma_iters))

            if sigma is None:
                return

            input_data = sigma_clip(input_data, sigma=sigma, sigma_lower=sigma_lower,
                                       sigma_upper=sigma_upper, iters=sigma_iters, axis=0)

            # Add to label so it is clear which overlay/component is which
            if sigma:
                label += ' sigma={}'.format(sigma)

            if sigma_lower:
                label += ' sigma_lower={}'.format(sigma_lower)

            if sigma_upper:
                label += ' sigma_upper={}'.format(sigma_upper)

            if sigma_iters:
                label += ' sigma_iters={}'.format(sigma_iters)
        else:
            input_data = input_data # noop

        # Do calculation if we got this far
        new_wavelengths, new_component = collapse_cube(input_data, data_name, self.data.coords.wcs,
                                             operation, start_index, end_index)

        return new_wavelengths, new_component, label

    def clear_stylesheets(self):
        self.ui.start_label.setStyleSheet("color: rgba(0, 0, 0, 255)")
        self.ui.end_label.setStyleSheet("color: rgba(0, 0, 0, 255)")

        self.ui.simple_sigma_label.setStyleSheet("color: rgba(0, 0, 0, 255)")

        self.ui.advanced_sigma_label.setStyleSheet("color: rgba(0, 0, 0, 255)")
        self.ui.advanced_sigma_lower_label.setStyleSheet("color: rgba(0, 0, 0, 255)")
        self.ui.advanced_sigma_upper_label.setStyleSheet("color: rgba(0, 0, 0, 255)")
        self.ui.advanced_sigma_iters_label.setStyleSheet("color: rgba(0, 0, 0, 255)")

    def calculate_callback(self):
        """
        Callback for when they hit calculate
        :return:
        """

        log.debug('In calculate_callback()')

        self.clear_stylesheets()

        # Grab the values of interest
        data_name = self.data_combobox.currentText()
        using_wavelengths = not 'Indices' in self.region_combobox.currentText()
        start_value = self.ui.start_input.text().strip()
        end_value = self.ui.end_input.text().strip()
        log.debug('    data_name {}  using_wavelengths {}  start_value {}  end_value {}'.format(
            data_name, using_wavelengths, start_value, end_value))

        # Clear the style sheet errors
        self.ui.error_label.setText(' ')
        self.ui.error_label.setStyleSheet("color: rgba(255, 0, 0, 128)")
        self.ui.error_label.setVisible(False)

        # Do check on wavelength/indices first.
        if using_wavelengths:
            start_index, end_index = self._calculate_callback_wavelength_checks(start_value, end_value)
        else:
            start_index, end_index = self._calculate_callback_index_checks(start_value, end_value)

        if start_index is None or end_index is None:
            return

        # Check to see if the wavelength (indices) are the same.
        if start_index == end_index:
            self.ui.start_label.setStyleSheet("color: rgba(255, 0, 0, 128)")
            self.ui.end_label.setStyleSheet("color: rgba(255, 0, 0, 128)")
            self.ui.error_label.setText('Can not have both start and end wavelengths be the same.')
            self.ui.error_label.setVisible(True)
            return

        data_name = self.data_combobox.currentText()
        operation = self.operation_combobox.currentText()
        spatial_region = self.spatial_region_combobox.currentText()
        sigma_selection = self.sigma_combobox.currentText()

        # Get the start and end wavelengths from the newly created
        # spectral cube and use for labeling the cube.
        # Convert to the current units.
        start_wavelength = self.wavelengths[start_index]
        end_wavelength = self.wavelengths[end_index]

        if 'Simple' in sigma_selection:
            sigma_parameter = self._calculate_callback_simple_sigma_check()
            if sigma_parameter is None:
                return
        elif 'Advanced' in sigma_selection:
            sigma_parameter = self._calculate_callback_advanced_sigma_check()
            if sigma_parameter[0] is None:
                return
        else:
            sigma_parameter = None

        # Do the actual call.
        new_wavelengths, new_component, label = self._calculate_collapse(
                data_name, operation, spatial_region,
                sigma_selection, sigma_parameter,
                start_index, end_index)

        # Add new overlay/component to cubeviz. We add this both to the 2D
        # container Data object and also as an overlay. In future we might be
        # able to use the 2D container Data object for the overlays directly.

        try:
            add_to_2d_container(self.parent, self.data, new_component, label)
            self.parent.add_overlay(new_component, label, display_now=False)
        except Exception as e:
            show_error_message(str(e), 'Collapse Cube Error', parent=self)
            return
        finally:
            self.close()

        # Show new dialog
        self.final_dialog(label)


    def final_dialog(self, label):
        """
        Final dialog that to show where the calculated collapsed cube was put.

        :param label:
        :return:
        """

        final_dialog = QDialog()

        # Create data component label and input box
        widget_desc = QLabel('The collapsed cube was added as an overlay with label "{}". {}'.format(
            label, self._extra_message))
        widget_desc.setWordWrap(True)
        widget_desc.setFixedWidth(350)
        widget_desc.setAlignment((Qt.AlignLeft | Qt.AlignTop))

        hb_desc = QHBoxLayout()
        hb_desc.addWidget(widget_desc)

        # Create Ok button
        okButton = QPushButton("Ok")
        okButton.clicked.connect(lambda: final_dialog.close())
        okButton.setDefault(True)

        hb_buttons = QHBoxLayout()
        hb_buttons.addStretch(1)
        hb_buttons.addWidget(okButton)

        # Add description and buttons to popup box
        vbl = QVBoxLayout()
        vbl.addLayout(hb_desc)
        vbl.addLayout(hb_buttons)

        final_dialog.setLayout(vbl)
        final_dialog.setMaximumWidth(400)
        final_dialog.show()

    def cancel_callback(self, caller=0):
        """
        Cancel callback when the person hits the cancel button

        :param caller:
        :return:
        """
        self.close()

    def keyPressEvent(self, e):
        if e.key() == Qt.Key_Escape:
            self.cancel_callback()


def collapse_cube(data_component, data_name, wcs, operation, start_index, end_index):
    """

    :param data_component:  Component from the data object
    :param wcs:
    :param operation:
    :param start:
    :param end:
    :return:
    """

    # Grab spectral-cube
    import spectral_cube

    # Create a spectral cube instance
    cube = spectral_cube.SpectralCube(data_component, wcs=wcs)

    # Do collapsing of the cube
    sub_cube = cube[start_index:end_index]
    calculated = sub_cube.apply_numpy_function(operations[operation], axis=0)

    wavelengths = sub_cube.spectral_axis

    # Send collapsed cube back to cubeviz
    return wavelengths, calculated



# -*- coding: utf-8 -*-
# This file contains a sub-class of the glue image viewer with further
# customizations.

import numpy as np

from matplotlib.axes import Axes
import matplotlib.image as mimage
from matplotlib.patches import Circle

from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.wcs.utils import wcs_to_celestial_frame
from astropy.coordinates import BaseRADecFrame

from qtpy.QtWidgets import (QLabel, QMessageBox)

from glue.core.message import SettingsChangeMessage

from glue.utils.qt import pick_item, get_text

from glue.viewers.image.qt import ImageViewer
from glue.viewers.image.layer_artist import ImageLayerArtist
from glue.viewers.image.state import ImageLayerState, ImageViewerState
from glue.viewers.image.qt.layer_style_editor import ImageLayerStyleEditor
from glue.viewers.common.qt.tool import Tool
from glue.viewers.matplotlib.state import DeferredDrawCallbackProperty as DDCProperty

from qtpy.QtWidgets import QToolTip
from qtpy.QtGui import QCursor

from .messages import (SliceIndexUpdateMessage, WavelengthUpdateMessage,
                       WavelengthUnitUpdateMessage, FluxUnitsUpdateMessage)
from .utils.contour import ContourSettings

CONTOUR_DEFAULT_NUMBER_OF_LEVELS = 8
CONTOUR_MAX_NUMBER_OF_LEVELS = 1000

__all__ = ['CubevizImageViewer']


def only_draw_axes_images(ax):
    """
    This function is a modified version of
    matplotlib.axes._base._AxesBase.draw
    This version only updates the images
    of the axes.
    :param ax: axes with images to update
    """
    if not ax.get_visible():
        return

    renderer = ax._cachedRenderer
    renderer.open_group('axes')

    # prevent triggering call backs during the draw process
    ax._stale = True

    locator = ax.get_axes_locator()
    if locator:
        pos = locator(ax, renderer)
        ax.apply_aspect(pos)
    else:
        ax.apply_aspect()

    # This is the biggest modification
    # Restrict the artists list to images
    artists = ax.images
    artists = sorted(artists, key=lambda x: x.zorder)

    # rasterize artists with negative zorder
    # if the minimum zorder is negative, start rasterization
    rasterization_zorder = ax._rasterization_zorder
    if (rasterization_zorder is not None and
            artists and artists[0].zorder < rasterization_zorder):
        renderer.start_rasterizing()
        artists_rasterized = [a for a in artists
                              if a.zorder < rasterization_zorder]
        artists = [a for a in artists
                   if a.zorder >= rasterization_zorder]
    else:
        artists_rasterized = []

    if artists_rasterized:
        for a in artists_rasterized:
            a.draw(renderer)
        renderer.stop_rasterizing()

    # This function will draw the images in the artist list
    mimage._draw_list_compositing_images(renderer, ax, artists)

    if hasattr(ax, "coords"):
        ax.coords.frame.draw(renderer)

    renderer.close_group('axes')
    ax._cachedRenderer = renderer
    ax.stale = False


class CubevizImageViewerState(ImageViewerState):

    # Override and modify ImageViewerState, so as to override the slice index
    # if needed.

    _slice_index_override = None

    @property
    def slice_index_override(self):
        return self._slice_index_override

    @slice_index_override.setter
    def slice_index_override(self, value):
        # The image viewer state uses smart caching to only update the data
        # when needed - some of which relies on checking for changes in the
        # slices - we therefore need to tell it about the change in the override
        slices_before = self.numpy_slice_aggregation_transpose[0]
        self._slice_index_override = value
        slices_after = self.numpy_slice_aggregation_transpose[0]
        for layer in self.layers:
            layer.reset_cache_from_slices(slices_before, slices_after)

    @property
    def numpy_slice_aggregation_transpose(self):
        slices, agg_func, transpose = super(CubevizImageViewerState, self).numpy_slice_aggregation_transpose
        if self.slice_index_override is not None:
            slices[0] = self.slice_index_override
        return slices, agg_func, transpose


class CubevizImageLayerState(ImageLayerState):
    """
    Sub-class of ImageLayerState that includes the ability to include smoothing
    on-the-fly.
    """

    preview_function = None

    # Override glue default
    global_sync = DDCProperty(False)

    def get_sliced_data(self, view=None):
        """
        Override and modify ImageLayerState.get_sliced_data so that if
        CubevizImageLayerState.preview_function is defined, it is applied to the
        data before return.
        """
        self._cache = None
        if self.preview_function is None:
            return super(CubevizImageLayerState, self).get_sliced_data(view=view)
        else:
            image = super(CubevizImageLayerState, self).get_sliced_data()
            image = self.preview_function(image)
            if view is not None:
                image = image[view]
            return image


class CubevizImageLayerStyleEditor(ImageLayerStyleEditor):

    # Override this in order to get rid of the sync button

    def __init__(self, *args, **kwargs):
        super(CubevizImageLayerStyleEditor, self).__init__(*args, **kwargs)
        self.ui.bool_global_sync.setVisible(False)


class CubevizImageLayerArtist(ImageLayerArtist):

    _state_cls = CubevizImageViewerState
    _layer_state_cls = CubevizImageLayerState


class CubevizImageViewer(ImageViewer):

    tools = ['select:rectangle', 'select:xrange', 'select:yrange',
             'select:circle', 'select:polygon', 'image:contrast_bias',
             'cubeviz:contour']

    _state_cls = CubevizImageViewerState
    _close_on_last_layer_removed = False

    def __init__(self,  *args, cubeviz_layout=None, **kwargs):
        super(CubevizImageViewer, self).__init__(*args, **kwargs)
        self.cubeviz_layout = cubeviz_layout
        self._layer_style_widget_cls[CubevizImageLayerArtist] = CubevizImageLayerStyleEditor
        self._hub = cubeviz_layout.session.hub
        self._synced_checkbox = None
        self._slice_index = None

        self.current_component_id = None  # Current component id

        self.cubeviz_unit = None
        self.component_unit_label = ""  # String to hold units of data values

        self.is_mouse_over = False  # If mouse cursor is over viewer
        self.hold_coords = False  # Switch to hold current displayed coords
        self._coords_in_degrees = False  # Switch display coords to True=deg or False=Deg/Hr:Min:Sec
        self._coords_format_function = self._format_to_hex_string  # Function to format ra and dec
        self.x_mouse = None  # x position of mouse in pix
        self.y_mouse = None  # y position of mouse in pix
        self.mouse_value = ""  # Value under mouse as string
        self._is_tooltip_on = True  # Display mouse_value as tool tip

        self.is_contour_active = False  # Is contour being displayed
        self.is_contour_preview_active = False # Is contour in preview mode
        self.contour = None  # matplotlib.axes.Axes.contour
        self.contour_component = None  # component label for contour
        self.contour_settings = ContourSettings(self)  # ContourSettings
        self.contour_preview_settings = None  # Temporary ContourSettings

        self.is_smoothing_preview_active = False  # Smoothing preview flag
        self.smoothing_preview_title = ""

        self.is_axes_hidden = False  # True if axes is hidden
        self.axes_title = ""  # Plot title

        self._has_2d_data = False  # True if currently displayed data is 2D
        self._toggle_3d = False  # True if we just toggled from 2D to 3D

        self._subset = None # Keep track of currently active subset

        self.coord_label = QLabel("")  # Coord display
        self.statusBar().addPermanentWidget(self.coord_label)

        # These are updated by listeners. See hub subscribers below
        self._wavelengths = None  # Array of wavelengths to display
        self._wavelength_units = None  # Units to use for wavelength values

        # Connect matplotlib events to event handlers
        self.statusBar().messageChanged.connect(self.message_changed_callback)
        self.figure.canvas.mpl_connect('motion_notify_event', self.mouse_move)
        self.figure.canvas.mpl_connect('figure_enter_event', self.turn_mouse_on)

        self._dont_update_status = False  # Don't save statusBar message when coords are changing
        self.status_message = self.statusBar().currentMessage()

        # Allow the CubeViz slider to respond to viewer-specific sliders in the glue pane
        self.state.add_callback('slices', self._slice_callback)

        self._hub.subscribe(self, SliceIndexUpdateMessage, handler=self._update_viewer_index)
        self._hub.subscribe(self, WavelengthUpdateMessage, handler=self._update_wavelengths)
        self._hub.subscribe(self, WavelengthUnitUpdateMessage, handler=self._update_wavelength_units)
        self._hub.subscribe(self, FluxUnitsUpdateMessage, handler=self._update_flux_units)

    def _slice_callback(self, new_slice):
        if self._slice_index is not None and not self.has_2d_data:
            self.cubeviz_layout._slice_controller.update_index(new_slice[0])
        # When toggling from 2D to 3D data component, update to synced index
        elif self._toggle_3d:
            self._toggle_3d = False
            self.cubeviz_layout._slice_controller.update_index(self.cubeviz_layout.synced_index)
            self.update_slice_index(self.cubeviz_layout.synced_index)

    def get_data_layer_artist(self, layer=None, layer_state=None):
        if layer.ndim == 1:
            cls = self._scatter_artist
        else:
            cls = CubevizImageLayerArtist
        return self.get_layer_artist(cls, layer=layer, layer_state=layer_state)

    def _update_stats_text(self, label, min_, max_, median, mu, sigma):
        text = r"min={:.4}, max={:.4}, median={:.4}, μ={:.4}, σ={:.4}".format(min_, max_, median, mu, sigma)
        self.parent().set_stats_text(label, text)

    def _calculate_stats(self, data):
        if self.cubeviz_unit is not None:
            wave = self.cubeviz_layout.get_wavelength(self.slice_index)
            data = self.cubeviz_unit.convert_from_original_unit(data, wave=wave)
        return np.nanmin(data), np.nanmax(data), np.median(data), data.mean(), data.std()

    def show_roi_stats(self, component, subset):

        if self._has_2d_data or subset.ndim != 3:
            self.parent().set_stats_text('', '')
            return

        self._subset = subset

        mask = subset.to_mask()[self._slice_index]
        data = self._data[0][component][self._slice_index][mask]

        results = self._calculate_stats(data)
        label = '{} Statistics:'.format(subset.label)
        self._update_stats_text(label, *results)

    def show_slice_stats(self):

        if self._has_2d_data:
            self.parent().set_stats_text('', '')
            return

        self._subset = None

        data = self._data[0][self.current_component_id][self._slice_index]
        results = self._calculate_stats(data.copy())
        self._update_stats_text('Slice Statistics:', *results)

    def update_stats(self):

        if self._subset is not None:
            self.show_roi_stats(self.current_component_id, self._subset)
        else:
            self.show_slice_stats()

    def update_component(self, component):
        self.update_stats()

    @property
    def is_preview_active(self):
        return self.is_contour_preview_active or self.is_smoothing_preview_active

    @property
    def has_2d_data(self):
        return self._has_2d_data or self._toggle_3d

    @has_2d_data.setter
    def has_2d_data(self, value):
        if self._has_2d_data and not value:
            self._toggle_3d = True
        self._has_2d_data = value

    def update_axes_title(self, title=None):
        """
        Update plot title.
        :param title: str: Plot title
        """
        if title is not None:
            if self.component_unit_label:
                self.axes_title = "{0} [{1}]".format(title, self.component_unit_label)
            else:
                self.axes_title = title

        if self.is_contour_preview_active:
            return

        self.axes.set_title(self.axes_title, color="black")
        self.axes.figure.canvas.draw()

        # Disabled feature:
        #self.statusBar().showMessage(self.axes_title)

    def show_preview_title(self):
        """
        Override normal plot title to show smoothing preview title.
        """
        if self.is_smoothing_preview_active:
            title = self.smoothing_preview_title
        elif self.is_contour_preview_active:
            title = "Contour Preview"
        else:
            return

        if self.is_axes_hidden:
            st = self.figure.suptitle(title, color="black")
            st.set_bbox(dict(facecolor='red', edgecolor='red'))
            self.axes.set_title("", color="black")
        else:
            self.axes.set_title(title, color="r")

    def hide_preview_title(self):
        """
        You should always call this after setting the is_preview flag
        """
        self.figure.suptitle("", color="black")
        if self.is_preview_active:
            self.show_preview_title()
        else:
            self.update_axes_title()

    def set_smoothing_preview(self, preview_function, preview_title=None):
        """
        Sets up on the fly smoothing and displays smoothing preview title.
        :param preview_function: function: Single-slice smoothing function
        :param preview_title: str: Title displayed when previewing
        """
        self.is_smoothing_preview_active = True

        if preview_title is None:
            self.smoothing_preview_title = "Smoothing Preview"
        else:
            self.smoothing_preview_title = preview_title
        self.show_preview_title()

        for layer in self.layers:
            if isinstance(layer, CubevizImageLayerArtist):
                layer.state.preview_function = preview_function
        self.axes._composite_image.invalidate_cache()
        if self.is_contour_active:
            self.draw_contour()
        else:
            self.axes.figure.canvas.draw()

    def end_smoothing_preview(self):
        """
        Ends on the fly smoothing.
        Warning: A change of combo index should always happen
        after calling this function!
        """
        self.is_smoothing_preview_active = False
        self.hide_preview_title()
        self.smoothing_preview_title = "Smoothing Preview"
        for layer in self.layers:
            if isinstance(layer, CubevizImageLayerArtist):
                layer.state.preview_function = None
        self.axes._composite_image.invalidate_cache()
        if self.is_contour_active:
            self.draw_contour()
        else:
            self.axes.figure.canvas.draw()

    def toggle_hidden_axes(self, is_axes_hidden):
        """
        Opertations to execute when axes is hidden/shown.
        :param is_axes_hidden: bool: True if axes is now hidden
        """
        self.is_axes_hidden = is_axes_hidden

        # Plot title operations
        if self.is_preview_active:
            self.show_preview_title()
        else:
            self.update_axes_title()

    def _delete_contour(self):
        if self.contour is not None:
            for c in self.contour.collections:
                c.remove()

            for c in self.contour.labelTexts:
                c.remove()
            del self.contour
            self.contour = None

    def get_contour_array(self):
        if self.contour_component is None:
            layer_artist = self.first_visible_layer()
            arr = layer_artist.state.get_sliced_data()
        else:
            data = self.state.layers_data[0]
            arr = data[self.contour_component][self.slice_index]

        if self.cubeviz_unit is not None:
            arr = arr.copy()
            wave = self.cubeviz_layout.get_wavelength(self.slice_index)
            arr = self.cubeviz_unit.convert_from_original_unit(arr, wave=wave)

        return arr

    def draw_contour(self, draw=True):
        self._delete_contour()

        if len(self.visible_layers()) == 0:
            return

        if self.is_contour_preview_active:
            settings = self.contour_preview_settings
        else:
            settings = self.contour_settings

        arr = self.get_contour_array()

        vmax = np.nanmax(arr)
        if settings.vmax is not None:
            vmax = settings.vmax

        vmin = np.nanmin(arr)
        if settings.vmin is not None:
            vmin = settings.vmin

        if settings.spacing is None:
            spacing = 1
            if vmax != vmin:
                spacing = (vmax - vmin)/CONTOUR_DEFAULT_NUMBER_OF_LEVELS
        else:
            spacing = settings.spacing

        levels = np.arange(vmin, vmax, spacing)
        levels = np.append(levels, vmax)

        if levels.size > CONTOUR_MAX_NUMBER_OF_LEVELS:
            message = "The current contour spacing is too small and " \
                      "results in too many levels. Contour spacing " \
                      "settings have been reset to auto."
            info = QMessageBox.critical(self, "Error", message)

            settings.spacing = None
            settings.data_spacing = spacing
            if settings.dialog is not None:
                settings.dialog.custom_spacing_checkBox.setChecked(False)
            spacing = (vmax - vmin)/CONTOUR_DEFAULT_NUMBER_OF_LEVELS
            levels = np.arange(vmin, vmax, spacing)
            levels = np.append(levels, vmax)

        self.contour = self.axes.contour(arr, levels=levels, **settings.options)

        if settings.add_contour_label:
            if abs(levels).max() > 1000 \
                    or 0.0 < abs(levels).min() < 0.001 \
                    or 0.0 < abs(levels).max() < 0.001:
                self.axes.clabel(self.contour,
                                 fmt='%.2E',
                                 fontsize=settings.font_size)
            else:
                self.axes.clabel(self.contour, fontsize=settings.font_size)

        settings.data_max = arr.max()
        settings.data_min = arr.min()
        settings.data_spacing = spacing
        if settings.dialog is not None:
            settings.update_dialog()
        if draw:
            self.axes.figure.canvas.draw()

    def default_contour(self, *args):
        """
        Draw contour of current component
        :param args: arguments from toolbar
        """
        self.is_contour_active = True
        self.contour_component = None
        self.draw_contour()

    def custom_contour(self, *args):
        """
        Draw contour of a specified component
        To change component programmatically
        change the `contour_component` class var
        :param args: arguments from toolbar
        """

        components = self.cubeviz_layout.component_labels
        self.contour_component = pick_item(components, components,
                                           title='Custom Contour',
                                           label='Pick a component')
        if self.contour_component is None:
            return
        else:
            self.is_contour_active = True
            self.draw_contour()

    def remove_contour(self, *args):
        """
        Turn contour off
        :param args: arguments from toolbar
        """
        self.is_contour_active = False
        self._delete_contour()
        self.axes.figure.canvas.draw()

    def edit_contour_settings(self, *args):
        """
        Edit contour settings using UI
        :param args: arguments from toolbar
        :return: settings UI
        """
        arr = self.get_contour_array()
        vmax = arr.max()
        vmin = arr.min()
        spacing = 1
        if vmax != vmin:
            spacing = (vmax - vmin)/CONTOUR_DEFAULT_NUMBER_OF_LEVELS
        self.contour_settings.data_max = vmax
        self.contour_settings.data_min = vmin
        self.contour_settings.data_spacing = spacing

        return self.contour_settings.options_dialog()

    def set_contour_preview(self, contour_preview_settings):
        """
        Apply contour preview settings.
        :param contour_preview_settings: ContourSettings
        """
        self.is_contour_preview_active = True
        self.show_preview_title()
        self.contour_preview_settings = contour_preview_settings
        self.draw_contour()

    def end_contour_preview(self):
        """ End contour preview"""
        self.is_contour_preview_active = False
        self.hide_preview_title()
        self.contour_preview_settings = None
        if self.is_contour_active:
            self.draw_contour()
        else:
            self.remove_contour()

    def _synced_checkbox_callback(self, event):
        if self._synced_checkbox.isChecked():
            msg = SettingsChangeMessage(self, [self])
            self.parent().tab_widget.session.hub.broadcast(msg)
            self.update_slice_index(self.parent().tab_widget.synced_index)

    def assign_synced_checkbox(self, checkbox):
        self._synced_checkbox = checkbox
        self._synced_checkbox.stateChanged.connect(self._synced_checkbox_callback)

    def _update_viewer_index(self, message):

        index = message.index
        active_cube = self.cubeviz_layout._active_cube
        active_widget = active_cube._widget

        # If the active widget is synced then we need to update the image
        # in all the other synced views.
        if (active_widget.synced and self.synced and not \
            self.cubeviz_layout._single_viewer_mode) or \
                active_widget is self:
            if message.slider_down:
                self.fast_draw_slice_at_index(index)
            else:
                self.update_slice_index(index)

        self.parent().slice_text.setText('slice: {:5}'.format(self._slice_index))

    def update_slice_index(self, index):
        """
        Function to update image and slice index.
        Redraws figure.
        :param index: (int) slice index
        """

        # Reset slice index override
        self.state.slice_index_override = None

        # Do not update if displaying a 2D data component
        if len(self.state.slices) == 2:
            return

        self._slice_index = index
        z, y, x = self.state.slices
        self.state.slices = (self._slice_index, y, x)
        if self.is_contour_active:
            self.draw_contour()

        self.update_stats()

    def fast_draw_slice_at_index(self, index):
        """
        Function to update the displayed image at a slice index
        quickly. Used when the user is scrolling using a slider.
        Utilizes a modified version of matplotlib's `axes.draw()`
        function to only draw images then uses `fig.canvas.blit()`
        to update the canvas.
        :param index: (int) slice index
        """

        if self._slice_index == index:
            return

        # draw_artist can only be used after an
        # initial draw which caches the render
        if self.axes._cachedRenderer is None:
            self.update_slice_index(index)
            return

        self._slice_index = index

        # Set main image's slice index to index
        self.state.slice_index_override = index

        # Invalidate cached data for image viewer
        # Or else it will not be redrawn
        self.axes._composite_image.invalidate_cache()

        # Redraw canvas images
        fig = self.axes.figure
        for ax in fig.axes:
            only_draw_axes_images(ax)

        # Draw contour and its labels
        ax = self.axes
        if self.is_contour_active and self.contour is not None:
            self.draw_contour(draw=False)
            for c in self.contour.collections:
                ax.draw_artist(c)
            for t in self.contour.labelTexts:
                ax.draw_artist(t)

        # update canvas using blit
        fig.canvas.blit()

    @property
    def synced(self):
        return self._synced_checkbox.isChecked() and not self.has_2d_data

    @synced.setter
    def synced(self, value):
        self._synced_checkbox.setChecked(value)

    @property
    def slice_index(self):
        return self._slice_index

    def _update_flux_units(self, message):
        target_component_id = message.component_id
        if str(self.current_component_id) == str(target_component_id):
            self.update_component_unit_label(target_component_id)
            self.update_axes_title(str(target_component_id))
            self.update_slice_index(self.slice_index)

    def update_component_unit_label(self, component_id=None):
        """
        Update component's unit label.
        :param component_id: component id
        """
        if component_id is None:
            if self.current_component_id is None:
                self.component_unit_label = ""
                return self.component_unit_label
            component_id = self.current_component_id
        data = component_id.parent
        unit = str(data.get_component(component_id).units)
        if unit:
            self.component_unit_label = "{0}".format(unit)
        else:
            self.component_unit_label = ""
        return self.component_unit_label

    def get_coords(self):
        """
        Returns coord display string.
        """
        if self.is_mouse_over:
            return self.coord_label.text()
        return None

    def toggle_hold_coords(self):
        """
        Switch hold_coords state
        """
        if self.hold_coords:
            self.statusBar().showMessage("")
            self.hold_coords = False
        else:
            self.statusBar().showMessage("Frozen Coordinates")
            self.hold_coords = True

    def init_ra_dec(self):
        """
        Initialize the format of RA and DEC
        on the image viewers.
        Options:
            1) Sexagesimal:
                _coords_in_degrees -> False
                formatter_0 -> 'hh:mm:ss'
                formatter_1 -> 'dd:mm:ss'
            2) Decimal Degrees:
                _coords_in_degrees -> True
                formatter_0 -> 'd.dddd'
                formatter_1 -> 'd.dddd'
        """
        self._coords_in_degrees = False
        self.axes.coords[0].set_major_formatter('hh:mm:ss.s')
        self.axes.coords[1].set_major_formatter('dd:mm:ss')
        self.figure.canvas.draw()

    def toggle_coords_in_degrees(self):
        """
        Switch coords_in_degrees state
        """
        data = self.state.layers_data[0]
        is_ra_dec = isinstance(wcs_to_celestial_frame(data.coords.wcs),
                               BaseRADecFrame)
        if self._coords_in_degrees:
            self._coords_in_degrees = False
            self._coords_format_function = self._format_to_hex_string
            if is_ra_dec:
                self.axes.coords[0].set_major_formatter('hh:mm:ss.s')
                self.axes.coords[1].set_major_formatter('dd:mm:ss')
                self.figure.canvas.draw()
        else:
            self._coords_in_degrees = True
            self._coords_format_function = self._format_to_degree_string
            if is_ra_dec:
                self.axes.coords[0].set_major_formatter('d.dddd')
                self.axes.coords[1].set_major_formatter('d.dddd')
                self.figure.canvas.draw()

    def message_changed_callback(self, event):
        """
        This will be used to swap tool messages and coords messages.
        When coords are displayed, tool message is cleared.
        So we save the tool message and update it when it changes.
        This callback is for when the tool message changes and the
        boolean associated is to ignore the tool messages from the
        coordinate display.
        :param event: str: New status bar message.
        """
        if self._dont_update_status:
            return
        self.status_message = event

    def clear_coords(self):
        """
        Reset coord display and mouse tracking variables.
        If hold_coords is active (True), make changes
        only to indicate that the mouse is no longer over viewer.
        """
        if self.hold_coords:
            return
        if QToolTip.text() == self.mouse_value:
            QToolTip.hideText()
        self.x_mouse = None
        self.y_mouse = None
        self.coord_label.setText('')
        self.statusBar().showMessage(self.status_message)

    def _update_wavelengths(self, message):
        self._wavelengths = message.wavelengths

    def _update_wavelength_units(self, message):
        self._wavelength_units = message.units

    def _format_to_degree_string(self, ra, dec):
        """
        Format RA and Dec in degree format. If wavelength
        is available add it to the output sting.
        :return: string
        """
        coord_string = "({:0>8.4f}, {:0>8.4f}".format(ra, dec)

        # Check if wavelength is available
        if self.slice_index is not None and self._wavelengths is not None:
            wave = self._wavelengths[self.slice_index]
            coord_string += ", {:1.2e}{})".format(wave, self._wavelength_units)
        else:
            coord_string += ")"

        return coord_string

    def _format_to_hex_string(self, ra, dec):
        """
        Format RA and Dec in D:M:S and H:M:S formats respectively. If wavelength
        is available add it to the output sting.
        :return: string
        """
        c = SkyCoord(ra=ra * u.degree, dec=dec * u.degree)
        coord_string = "("
        coord_string += "{0:0>2.0f}h:{1:0>2.0f}m:{2:0>2.0f}s".format(*c.ra.hms)
        coord_string += ", "
        coord_string += "{0:0>3.0f}d:{1:0>2.0f}m:{2:0>2.0f}s".format(*c.dec.dms)

        # Check if wavelength is available
        if self.slice_index is not None and self._wavelengths is not None:
            wave = self._wavelengths[self.slice_index]
            coord_string += ", {:1.2e}{})".format(wave, self._wavelength_units)
        else:
            coord_string += ")"

        return coord_string

    def mouse_move(self, event):
        """
        Event handler for matplotlib motion_notify_event.
        Updates coord display and vars.
        :param event: matplotlib event.
        """
        # Check if mouse is in widget but not on plot
        if not event.inaxes:
            self.is_mouse_over = False
            self.clear_coords()
            return
        mouse_pos = QCursor.pos()
        self.is_mouse_over = True

        # If hold_coords is active, return
        if self.hold_coords:
            return

        # Get x and y of the pixel under the mouse
        x, y = [int(event.xdata + 0.5), int(event.ydata + 0.5)]
        self.x_mouse, self.y_mouse = [x, y]

        # Create coord display string
        if self._slice_index is not None:
            string = "({:1.0f}, {:1.0f}, {:1.0f})".format(x, y, self._slice_index)
        else:
            string = "({:1.0f}, {:1.0f})".format(x, y)

        self.mouse_value = ""

        # If viewer has a layer.
        if len(self.visible_layers()) > 0:

            arr = self.first_visible_layer().state.get_sliced_data()

            if 0 <= y < arr.shape[0] and 0 <= x < arr.shape[1]:
                # if x and y are in bounds. Note: x and y are swapped in array.
                # get value and check if wcs is obtainable
                # WCS:
                if len(self.figure.axes) > 0:
                    wcs = self.figure.axes[0].wcs.celestial
                    if wcs is not None:
                        # Check the number of axes in the WCS and add to string
                        ra = dec = None
                        if wcs.naxis == 3 and self.slice_index is not None:
                            ra, dec, wave = wcs.wcs_pix2world([[x, y, self._slice_index]], 0)[0]
                        elif wcs.naxis == 2:
                            ra, dec = wcs.wcs_pix2world([[x, y]], 0)[0]

                        if ra is not None and dec is not None:
                            string = string + " " + self._coords_format_function(ra, dec)
                # Pixel Value:
                v = arr[y][x]
                if self.cubeviz_unit is not None:
                    wave = self.cubeviz_layout.get_wavelength(self.slice_index)
                    v = self.cubeviz_unit.convert_from_original_unit(v, wave=wave)

                unit_string = ""
                if self.component_unit_label:
                    unit_string = "[{0}] ".format(self.component_unit_label)

                if 0.01 <= abs(v) <= 1000 or abs(v) == 0.0:
                    value_string = "{0:.3f} ".format(v)
                else:
                    value_string = "{0:.3e} ".format(v)

                self.mouse_value = value_string + unit_string
                string = value_string + string
        # Add a gap to string and add to viewer.
        string += " "
        self._dont_update_status = True
        self.statusBar().clearMessage()
        self._dont_update_status = False
        self.coord_label.setText(string)

        if self._is_tooltip_on:
            if self.mouse_value:
                QToolTip.showText(mouse_pos, "...", self)
                QToolTip.showText(mouse_pos, self.mouse_value, self)
            else:
                QToolTip.showText(mouse_pos, "...", self)
                QToolTip.showText(mouse_pos, " ", self)

        return

    def first_visible_layer(self):
        layers = self.visible_layers()
        if len(layers) == 0:
            raise Exception("Couldn't find any visible layers")
        else:
            return layers[0]

    def visible_layers(self):
        layers = []
        for layer_artist in self.layers:
            if layer_artist.enabled and layer_artist.visible:
                layers.append(layer_artist)
        return layers

    def mouse_exited(self, event):
        """
        Event handler for matplotlib axes_leave_event.
        Clears coord display and vars.
        :param event: matplotlib event
        """
        self.clear_coords()
        return

    def leaveEvent(self, event):
        """
        Event handler for Qt widget leave events.
        Clears coord display and vars.
        Overrides default.
        :param event: QEvent
        """
        self.clear_coords()
        self.is_mouse_over = False
        return

    def turn_mouse_on(self, event):
        self.is_mouse_over = True



import os

import numpy as np
from glue.core import Subset, Data
from qtpy.QtCore import QThread, Signal, QEventLoop
from qtpy.QtWidgets import QDialog, QDialogButtonBox
from qtpy.uic import loadUi
from spectral_cube import BooleanArrayMask, SpectralCube

__all__ = ['SpectralOperationHandler']

UI_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                       '..', 'data', 'ui'))


class SpectralOperationHandler(QDialog):
    """
    Widget to handle user interactions with operations that are communicated
    from the SpecViz viewer. This is built to work with
    :func:`~spectral_cube.SpectralCube.apply_function` method by passing in a
    callable :class:`specviz.analysis.operations.FunctionalOperation` object.

    Attributes
    ----------
    data : :class:`~glue.core.data.Data`
        Glue data object on which the spectral operation will be performed.
    function : :class:`specviz.analysis.operations.FunctionalOperation`
        Python class instance whose `call` function will be performed on the
        :class:`~spectral_cube.SpectralCube` object.
    """

    def __init__(self, data, stack, session, parent, *args, **kwargs):
        super(SpectralOperationHandler, self).__init__(*args, **kwargs)
        self.data = data
        self.stack = stack
        self.function = stack[0] if len(stack) > 0 else None
        self.component_id = self.data.component_ids()[0]
        self._operation_thread = None
        self._session = session
        self._parent = parent

        self.setup_ui()
        self.setup_connections()

    def setup_ui(self):
        """Setup the PyQt UI for this dialog."""
        # Load the ui dialog
        loadUi(os.path.join(UI_PATH, "apply_operation.ui"), self)

        component_ids = [str(i) for i in self.data.component_ids()]

        # Populate combo box
        self.data_component_combo_box.addItems(component_ids)

        operation_stack = []

        for oper in self.stack:
            func_params = []

            for arg in oper.args:
                func_params.append("{}".format(arg))

            for k, v in oper.kwargs.items():
                func_params.append("{}={}".format(k, str(v)[:10] + "... " + str(v)[-5:] if len(str(v)) > 15 else str(v)))

            operation_stack.append("{}({})".format(oper.function.__name__, ", ".join(func_params)))

        self.operation_combo_box.addItems(operation_stack)

        # Disable the button box if there are no available operations
        if self.function is None:
            self.button_box.button(QDialogButtonBox.Ok).setEnabled(False)

    def setup_connections(self):
        """Setup signal/slot connections for this dialog."""
        # When an operation is selected, update the function reference
        self.operation_combo_box.currentIndexChanged.connect(
            self.on_operation_index_changed)

        # When a data component is selected, update the data object reference
        self.data_component_combo_box.currentIndexChanged.connect(
            self.on_data_component_index_changed)

        # If the abort button is clicked, attempted to stop execution
        self.abort_button.clicked.connect(self.on_aborted)

    def _compose_cube(self):
        """
        Create a :class:`~spectral_cube.SpectralCube` from a Glue data
        component.
        """
        if issubclass(self.data.__class__, Subset):
            wcs = self.data.data.coords.wcs
            data = self.data.data
            mask = self.data.to_mask()
        else:
            wcs = self.data.coords.wcs
            data = self.data
            mask = np.ones(data.shape).astype(bool)

        mask = BooleanArrayMask(mask=mask, wcs=wcs)

        return SpectralCube(data[self.component_id], wcs=wcs, mask=mask)

    def on_operation_index_changed(self, index):
        """Called when the index of the operation combo box has changed."""
        self.function = self.stack[index]

    def on_data_component_index_changed(self, index):
        """Called when the index of the component combo box has changed."""
        self.component_id = self.data.component_ids()[index]

    def accept(self):
        """Called when the user clicks the "Okay" button of the dialog."""
        # Show the progress bar and abort button
        self.progress_bar.setEnabled(True)
        self.abort_button.setEnabled(True)

        self.button_box.button(QDialogButtonBox.Ok).setEnabled(False)
        self.button_box.button(QDialogButtonBox.Cancel).setEnabled(False)

        self._operation_thread = OperationThread(self._compose_cube(),
                                                 function=self.function)

        self._operation_thread.finished.connect(self.on_finished)
        self._operation_thread.status.connect(self.on_status_updated)
        self._operation_thread.start()

    def on_aborted(self):
        """Called when the user aborts the operation."""
        self._operation_thread.abort()
        self.progress_bar.setValue(0)

        # Hide the progress bar and abort button
        self.progress_bar.hide()
        self.abort_button.setEnabled(False)

    def on_status_updated(self, value):
        """
        Called when the status of the operation has been updated. This can be
        optionally be passed a value to use as the new progress bar value.

        Attributes
        ----------
        value : float
            The value passed to the :class:`~qtpy.QtWidgets.QProgressBar`
            instance.
        """
        self.progress_bar.setValue(value * 100)

    def on_finished(self, data):
        """
        Called when the `QThread` has finished performing the operation on the
        `SpectralCube` object.

        Attributes
        ----------
        data : ndarray
            The result of the operation performed on the `SpectralCube` object.
        """
        if self.function.axis == 'cube':
            data = Data(x=data)
            self._session.data_collection.append(data)

            super(SpectralOperationHandler, self).accept()

        component_name = '{} spectral {}'.format(self.component_id,
                self.operation_combo_box.currentText())

        comp_count = len([x for x in self.data.component_ids()
                          if component_name in str(x)])

        if comp_count > 0:
            component_name = "{} {}".format(component_name, comp_count)

        if len(data.shape) < len(self.data.shape):
            self._parent.add_overlay(data, component_name)
        else:
            self.data.add_component(data, component_name)

        super(SpectralOperationHandler, self).accept()


class SimpleProgressTracker():
    """
    Simple container object to track the progress of an operation occuring in a
    :class:`~qtpyt.QtCore.QThread` instance. It is designed to be passed to
    :class:`~spectral_cube.SpectralCube` object to be called while performing
    operations.

    Attributes
    ----------
    total_value : float
        The maximum value of the progress.
    """

    def __init__(self, total_value):
        self._current_value = 0.0
        self._total_value = total_value
        self._abort_flag = False

    def __call__(self, value=None):
        self._current_value = value or self._current_value + 1

        if self._abort_flag:
            raise Exception("Process aborted.")

    @property
    def percent_value(self):
        """Return the completion amount as a percentage."""
        return self._current_value / self._total_value

    def abort(self):
        """
        Set the abort flag which will raise an error causing the operation
        to return immediately.
        """
        self._abort_flag = True


class OperationThread(QThread):
    """
    Thread in which an operation is performed on some
    :class:`~spectral_cube.SpectralCube` object to ensure that the UI does not
    freeze while the operation is running.

    Attributes
    ----------
    cube_data : :class:`~spectral_cube.SpectralCube`
        The cube data on which the operation will be performed.
    function : callable
        The function-like callable used to perform the operation on the cube.
    """
    status = Signal(float)
    finished = Signal(object)

    def __init__(self, cube_data, function, parent=None):
        super(OperationThread, self).__init__(parent)
        self._cube_data = cube_data
        self._function = function
        self._tracker = None
        self._axes =  {
            'cube': None,
            'spectral': self._cube_data.spectral_axis
        }

    def on_query_result(self, state):
        self.block.emit(state)

    def run(self):
        """Run the thread."""
        self._tracker = SimpleProgressTracker(
            self._cube_data.shape[1] * self._cube_data.shape[2])

        def progress_wrapper():
            self._tracker()
            self.status.emit(self._tracker.percent_value)

        if self._function.axis == 'cube':
            new_data = self._function(self._cube_data)
        else:
            new_data = self._cube_data.apply_function(
                self._function,
                spectral_axis=self._cube_data.spectral_axis,
                axis=0,
                keep_shape=self._function.keep_shape,
                update_function=progress_wrapper)

        self.finished.emit(new_data)

    def abort(self):
        """
        Abort the operation. Halts and returns immediately by raising an
        error.
        """
        self._tracker.abort()



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from glue.core import Hub, HubListener, Data, DataCollection
from glue.core.message import (DataCollectionAddMessage, SettingsChangeMessage,
                               DataRemoveComponentMessage, SubsetMessage,
                               EditSubsetMessage, DataAddComponentMessage)
from .layout import CubeVizLayout
from .messages import FluxUnitsUpdateMessage


CUBEVIZ_LAYOUT = 'cubeviz_layout'


class CubevizManager(HubListener):

    def __init__(self, session):
        self._session = session
        self._hub = session.hub
        self._app = session.application
        self._layout = None

        self._empty_layout = self._app.add_fixed_layout_tab(CubeVizLayout)
        self._app.close_tab(0, warn=False)

        # For cubeviz, ROI selection should be in 'NewMode' by default
        self._app._mode_toolbar.set_mode('new')

        self._hub.subscribe(
            self, DataCollectionAddMessage, handler=self.handle_new_dataset)
        self._hub.subscribe(
            self, DataAddComponentMessage, handler=self.handle_new_component)
        self._hub.subscribe(
            self, SettingsChangeMessage, handler=self.handle_settings_change)
        self._hub.subscribe(
            self, DataRemoveComponentMessage, handler=self.handle_remove_component)
        self._hub.subscribe(
            self, SubsetMessage, handler=self.handle_subset_message)
        self._hub.subscribe(
            self, EditSubsetMessage, handler=self.handle_subset_message)
        self._hub.subscribe(
            self, FluxUnitsUpdateMessage, handler=self.handle_flux_units_update)

        # Look for any cube data files that were loaded from the command line
        for data in session.data_collection:
            if data.meta.get(CUBEVIZ_LAYOUT, ''):
                self.configure_layout(data)

    def handle_new_dataset(self, message):
        data = message.data
        if data.meta.get(CUBEVIZ_LAYOUT, ''):
            self.configure_layout(data)

    def configure_layout(self, data):
        # Assume for now the data is not yet in any tab
        if self._empty_layout is not None:
            cubeviz_layout = self._empty_layout
        else:
            cubeviz_layout = self._app.add_fixed_layout_tab(CubeVizLayout)

        try:
            self.setup_data(cubeviz_layout, data)
        finally:
            self._empty_layout = None

    def handle_new_component(self, message):
        component_id = message.component_id
        data = component_id.parent
        if data is self._layout._flux_unit_controller.data:
            units = data.get_component(component_id).units
            self._layout._flux_unit_controller.add_component_unit(component_id, units)
        self._layout.display_component(component_id)

    def handle_remove_component(self, message):
        self._layout.remove_data_component(message.component_id)

    def handle_settings_change(self, message):
        if self._layout is not None:
            self._layout.handle_settings_change(message)

    def handle_subset_message(self, message):
        if self._layout is not None:
            self._layout.handle_subset_action(message)

    def handle_flux_units_update(self, message):
        if self._layout is not None:
            self._layout.refresh_flux_units(message)

    def hide_sidebar(self):
        self._app._ui.main_splitter.setSizes([0, 300])

    def setup_data(self, cubeviz_layout, data):
        # Automatically add data to viewers and set attribute for split viewers
        image_viewers = [cubeviz_layout.single_view._widget,
                         cubeviz_layout.split_views[0]._widget,
                         cubeviz_layout.split_views[1]._widget,
                         cubeviz_layout.split_views[2]._widget]

        data_headers = [str(x).strip() for x in data.component_ids() if not x in data.coordinate_components]

        # Single viewer should display FLUX only by default
        image_viewers[0].add_data(data)
        image_viewers[0].state.aspect = 'auto'
        image_viewers[0].state.layers[0].attribute = data.id[data_headers[0]]

        # Split image viewers should each show different component by default
        for ii, dh in enumerate(data_headers[:3]):
            view = image_viewers[1+ii]
            view.add_data(data)
            view.state.aspect = 'auto'
            view.state.layers[0].attribute = data.id[dh]

        # And then for the "other" viewers, load the data up
        for jj in range(ii+1,3):
            view = image_viewers[1+jj]
            view.add_data(data)
            view.state.aspect = 'auto'
            view.state.layers[0].attribute = data.id[dh]

        cubeviz_layout.add_data(data)

        index = self._app.get_tab_index(cubeviz_layout)
        self._app.tab_bar.rename_tab(index, "CubeViz: {}".format(data.label))

        self._layout = cubeviz_layout



from __future__ import absolute_import, division, print_function

from qtpy.QtWidgets import QMessageBox

from glue.core import Data
from glue.core.link_helpers import LinkSame
from glue.core.coordinates import WCSCoordinates


def add_to_2d_container(cubeviz_layout, data, component_data, label):
    """
    Given the cubeviz layout, a data object, a new 2D layer and a label, add
    the 2D layer to the data object and update the cubeviz layout accordingly.
    This creates the 2D container dataset if needed.
    """

    # If the 2D container doesn't exist, we create it here. This container is
    # basically just a Data object but we keep it in an attribute
    # ``container_2d`` on its parent dataset.
    if getattr(data, 'container_2d', None) is None:

        # For now, we assume that the 2D maps are always computed along the
        # spectral axis, so that the resulting WCS is always celestial
        coords = WCSCoordinates(wcs=data.coords.wcs.celestial)
        data.container_2d = Data(label=data.label + " [2d]", coords=coords)

        data.container_2d.add_component(component_data, label)

        cubeviz_layout.session.data_collection.append(data.container_2d)

        # NOTE: the following is disabled for now but can be uncommented once
        # we are ready to use the glue overlay infrastructure.
        # Set up pixel links so that selections in the image plane propagate
        # between 1D and 2D views. Again this assumes as above that the
        # moments are computed along the spectral axis
        # link1 = LinkSame(data.pixel_component_ids[2],
        #                  data.container_2d.pixel_component_ids[1])
        # link2 = LinkSame(data.pixel_component_ids[1],
        #                  data.container_2d.pixel_component_ids[0])
        # cubeviz_layout.session.data_collection.add_link(link1)
        # cubeviz_layout.session.data_collection.add_link(link2)

        for helper in cubeviz_layout._viewer_combo_helpers:
            helper.append_data(data.container_2d)

        for viewer in cubeviz_layout.cube_views:
            viewer._widget.add_data(data.container_2d)

    else:
        # Make sure we don't add duplicate data components
        if label in data.container_2d.component_ids():
            raise ValueError("Data component with label '{}' already exists, "
                             "and cannot be created again".format(label))

        data.container_2d.add_component(component_data, label)

def show_error_message(message, title, parent=None):

    box = QMessageBox(parent=parent)
    box.setIcon(QMessageBox.Warning)
    box.setText(message)
    box.setWindowTitle(title)
    box.setStandardButtons(QMessageBox.Ok)
    box.exec_()



from __future__ import absolute_import, division, print_function

from qtpy.QtCore import Qt
from qtpy import QtGui
from qtpy.QtWidgets import (QDialog, QComboBox, QPushButton,
                            QLabel, QWidget, QHBoxLayout, QVBoxLayout)

from .common import add_to_2d_container, show_error_message


# TODO: In the future, it might be nice to be able to work across data_collection elements

class MomentMapsGUI(QDialog):
    def __init__(self, data, data_collection, parent=None):
        super(MomentMapsGUI, self).__init__(parent)

        # Get the data_components (e.g., FLUX, DQ, ERROR etc)
        # Using list comprehension to keep the order of the component_ids
        self.data_components = [str(x).strip() for x in data.component_ids() if not x in data.coordinate_components]

        self.data = data
        self.data_collection = data_collection
        self.parent = parent

        self.label = ''

    def display(self):
        """
        Create the popup box with the calculation input area and buttons.

        :return:
        """
        self.setWindowFlags(self.windowFlags() | Qt.Tool)
        self.setWindowTitle("Create Moment Map")

        boldFont = QtGui.QFont()
        boldFont.setBold(True)

        # Create calculation label and input box
        self.data_label = QLabel("Data:")
        self.data_label.setFixedWidth(100)
        self.data_label.setAlignment((Qt.AlignRight | Qt.AlignTop))
        self.data_label.setFont(boldFont)

        self.data_combobox = QComboBox()
        self.data_combobox.addItems([str(x).strip() for x in self.data.component_ids() if not x in self.data.coordinate_components])
        self.data_combobox.setMinimumWidth(200)

        hbl1 = QHBoxLayout()
        hbl1.addWidget(self.data_label)
        hbl1.addWidget(self.data_combobox)

        # Create calculation label and input box
        self.order_label = QLabel("Order:")
        self.order_label.setFixedWidth(100)
        self.order_label.setAlignment((Qt.AlignRight | Qt.AlignTop))
        self.order_label.setFont(boldFont)

        self.order_combobox = QComboBox()
        self.order_combobox.addItems(["1", "2", "3", "4", "5", "6", "7", "8"])
        self.order_combobox.setMinimumWidth(200)

        hbl2 = QHBoxLayout()
        hbl2.addWidget(self.order_label)
        hbl2.addWidget(self.order_combobox)

        # Create Calculate and Cancel buttons
        self.calculateButton = QPushButton("Calculate")
        self.calculateButton.clicked.connect(self.calculate_callback)
        self.calculateButton.setDefault(True)

        self.cancelButton = QPushButton("Cancel")
        self.cancelButton.clicked.connect(self.cancel_callback)

        hbl5 = QHBoxLayout()
        hbl5.addStretch(1)
        hbl5.addWidget(self.cancelButton)
        hbl5.addWidget(self.calculateButton)

        # Add calculation and buttons to popup box
        vbl = QVBoxLayout()
        vbl.addLayout(hbl1)
        vbl.addLayout(hbl2)
        vbl.addLayout(hbl5)

        self.setLayout(vbl)
        self.setMaximumWidth(700)
        self.show()

    def do_calculation(self, order, data_name):
        # Grab spectral-cube
        import spectral_cube
        cube = spectral_cube.SpectralCube(self.data[data_name], wcs=self.data.coords.wcs)

        cube_moment = cube.moment(order=order, axis=0)

        self.label = '{}-moment-{}'.format(data_name, order)

        # Add new overlay/component to cubeviz. We add this both to the 2D
        # container Data object and also as an overlay. In future we might be
        # able to use the 2D container Data object for the overlays directly.
        add_to_2d_container(self.parent, self.data, cube_moment.value, self.label)
        self.parent.add_overlay(cube_moment.value, self.label, display_now=False)

    def calculate_callback(self):
        """
        Callback for when they hit calculate
        :return:
        """

        # Determine the data component and order
        order = int(self.order_combobox.currentText())
        data_name = self.data_combobox.currentText()

        try:
            self.do_calculation(order, data_name)
        except Exception as e:
            print('Error: {}'.format(e))
            show_error_message(str(e), 'Moment Map Error', parent=self)

        self.close()

    def cancel_callback(self, caller=0):
        """
        Cancel callback when the person hits the cancel button

        :param caller:
        :return:
        """
        self.close()

    def keyPressEvent(self, e):
        if e.key() == Qt.Key_Escape:
            self.cancel_callback()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import sys
import os
import argparse
import os

from glue.app.qt import GlueApplication
from glue.main import restore_session, get_splash, load_data_files, load_plugins
from qtpy.QtCore import QTimer
from qtpy import QtGui, QtWidgets

try:
    from glue.utils.qt.decorators import die_on_error
except ImportError:
    from glue.utils.decorators import die_on_error

from .version import version as cubeviz_version
from .data_factories import DataFactoryConfiguration

CUBEVIZ_ICON_PATH = os.path.abspath(
    os.path.join(
        os.path.abspath(__file__),
        "..",
        "data",
        "resources",
        "cubeviz_icon.png"
    )
)

CUBEVIZ_LOGO_PATH = os.path.abspath(
    os.path.join(
        os.path.abspath(__file__),
        "..",
        "data",
        "resources",
        "cubeviz_logo.png"
    )
)

def setup():

    from . import layout  # noqa
    from . import startup  # noqa


@die_on_error("Error starting up Cubeviz")
def main(argv=sys.argv):
    """
    The majority of the code in this function was taken from start_glue() in main.py after a discussion with
    Tom Robataille. We wanted the ability to get command line arguments and use them in here and this seemed
    to be the cleanest way to do it.

    :param argv:
    :return:
    """

    # # Parse the arguments, ignore any unkonwn
    parser = argparse.ArgumentParser()
    parser.add_argument("--data-configs", help="Directory or file for data configuration YAML files", action='append', default=[])
    parser.add_argument("--data-configs-show", help="Show the matching info", action="store_true", default=False)
    parser.add_argument('data_files', nargs=argparse.REMAINDER)
    args = parser.parse_known_args(argv[1:])

    # Store the args for each ' --data-configs' found on the commandline
    data_configs = args[0].data_configs
    data_configs_show = args[0].data_configs_show

    import glue
    from glue.utils.qt import get_qapp
    app = get_qapp()

    # Splash screen
    splash = get_splash()
    splash.image = QtGui.QPixmap(CUBEVIZ_LOGO_PATH)
    splash.show()

    # Start off by loading plugins. We need to do this before restoring
    # the session or loading the configuration since these may use existing
    # plugins.
    load_plugins(splash=splash)

    # Load the
    DataFactoryConfiguration(data_configs, data_configs_show, remove_defaults=True)

    datafiles = args[0].data_files

    # Check to make sure each file exists and raise an Exception
    # that will show in the popup if it does not exist.
    for fileparam in datafiles:
        for filename in  fileparam.split(','):
            if not os.path.isfile(filename.strip()):
                raise IOError('The file {} does not exist'.format(filename))

    # Show the splash screen for 1 second
    timer = QTimer()
    timer.setInterval(1000)
    timer.setSingleShot(True)
    timer.timeout.connect(splash.close)
    timer.start()

    data_collection = glue.core.DataCollection()
    hub = data_collection.hub

    splash.set_progress(100)

    session = glue.core.Session(data_collection=data_collection, hub=hub)
    ga = GlueApplication(session=session)
    ga.setWindowTitle('cubeviz ({})'.format(cubeviz_version))
    qapp = QtWidgets.QApplication.instance()
    qapp.setWindowIcon(QtGui.QIcon(CUBEVIZ_ICON_PATH))
    ga.setWindowIcon(QtGui.QIcon(CUBEVIZ_ICON_PATH))

    ga.run_startup_action('cubeviz')

    # Load the data files.
    if datafiles:
        datasets = load_data_files(datafiles)
        ga.add_datasets(data_collection, datasets, auto_merge=False)


    sys.exit(ga.start(maximized=True))



from __future__ import print_function, division

from glue.config import startup_action
from .listener import CubevizManager


_manager = None


@startup_action('cubeviz')
def cubeviz_setup(session, data_collection):
    global _manager
    _manager = CubevizManager(session)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import os

import pytest
import numpy as np

from qtpy import QtCore
from glue.core import roi
from cubeviz.tools.moment_maps import MomentMapsGUI

from ...tests.helpers import (toggle_viewer, select_viewer, left_click,
                      left_button_press, right_button_press, enter_slice_text,
                      assert_all_viewer_indices, assert_slice_text)


DATA_LABELS = ['018.DATA', '018.NOISE']


@pytest.fixture(scope='module')
def moment_maps(cubeviz_layout):
    cl = cubeviz_layout

    mm = MomentMapsGUI(cl._data, cl.session.data_collection, parent=cl)

    return mm


def assert_red_stylesheet(widget):
    assert widget.styleSheet() == "color: rgba(255, 0, 0, 128)"


def test_moment_maps_1(cubeviz_layout):
    # Create GUI
    mm = moment_maps(cubeviz_layout)
    mm.display()
    mm.order_combobox.setCurrentIndex(0)
    mm.data_combobox.setCurrentIndex(0)

    # Call calculate function and get result
    mm.calculate_callback()
    moment_component_id = [str(x) for x in cubeviz_layout._data.container_2d.component_ids() if str(x).startswith('018.DATA-moment-1')][0]
    np_result = cubeviz_layout._data.container_2d[moment_component_id]

    # Expected result
    np_data = cubeviz_layout._data[DATA_LABELS[0]]
    import spectral_cube
    cube = spectral_cube.SpectralCube(np_data, wcs=cubeviz_layout._data.coords.wcs)
    order = int(mm.order_combobox.currentText())
    cube_moment = np.asarray(cube.moment(order=order, axis=0))

    assert np.allclose(cube_moment, np_result, atol=1.0, equal_nan=True)


def test_moment_maps_2(cubeviz_layout):
    # Create GUI
    mm = moment_maps(cubeviz_layout)
    mm.display()
    mm.order_combobox.setCurrentIndex(1)
    mm.data_combobox.setCurrentIndex(0)

    # Call calculate function and get result
    mm.calculate_callback()
    moment_component_id = [str(x) for x in cubeviz_layout._data.container_2d.component_ids() if str(x).startswith('018.DATA-moment-2')][0]
    np_result = cubeviz_layout._data.container_2d[moment_component_id]

    # Expected result
    np_data = cubeviz_layout._data[DATA_LABELS[0]]
    import spectral_cube
    cube = spectral_cube.SpectralCube(np_data, wcs=cubeviz_layout._data.coords.wcs)
    order = int(mm.order_combobox.currentText())
    cube_moment = np.asarray(cube.moment(order=order, axis=0))

    assert np.allclose(cube_moment, np_result, atol=1.0, equal_nan=True)
